# copyright 2016-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda unit tests for XSD profile generation.

You may want to set the TEST_WRITE_SEDA_FILES environment variable to activate
writing of generated content back to the file-system.
"""

from doctest import Example
from itertools import chain, izip, repeat
import os
from os.path import basename, join

from six import binary_type, text_type

from lxml import etree
from lxml.doctestcompare import LXMLOutputChecker

from logilab.common.decorators import cached

from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_seda.xsd2yams import XSDMMapping
from cubicweb_seda.entities.profile_generation import _path_target_values, eid2xmlid

from cubicweb_seda import testutils


class XmlTestMixin(object):
    """Mixin class providing additional assertion methods for checking XML data."""
    NAMESPACES = {
        'xs': 'http://www.w3.org/2001/XMLSchema',
        'seda': 'fr:gouv:culture:archivesdefrance:seda:v2.0',
        'rng': 'http://relaxng.org/ns/structure/1.0',
        'a': 'http://relaxng.org/ns/compatibility/annotations/1.0',
    }
    # TBD in concret implementation
    schema_class = None
    schema_file = None
    adapter_id = None

    @classmethod
    @cached
    def schema(cls, xsd_filename):
        with open(xsd_filename) as stream:
            return cls.schema_class(etree.parse(stream))

    def xpath(self, element, expression):
        return element.xpath(expression, namespaces=self.NAMESPACES)

    def get_element(self, profile, name):
        """Return etree element definition for 'name' (there should be only one)"""
        elements = self.get_elements(profile, name)
        self.assertEqual(len(elements), 1)
        return elements[0]

    def get_elements(self, profile, name):
        """Return etree element definitions for 'name'"""
        raise NotImplementedError()

    def get_attribute(self, profile, name):
        """Return etree attribute definition for 'name' (there should be only one)"""
        attributes = self.get_attributes(profile, name)
        self.assertEqual(len(attributes), 1)
        return attributes[0]

    def get_attributes(self, profile, name):
        """Return etree attribute definitions for 'name'"""
        raise NotImplementedError()

    def profile_etree(self, transfer_entity, adapter_id=None):
        """Return etree representation of profile's XSD for the given transfer entity."""
        adapter = transfer_entity.cw_adapt_to(adapter_id or self.adapter_id)
        root = adapter.dump_etree()
        self.assertXmlValid(root)
        return root

    def assertXmlValid(self, root):
        """Validate an XML etree according to an XSD file (.xsd)."""
        schema = self.schema(self.datapath(self.schema_file))
        schema.assert_(root)

    def assertXmlEqual(self, actual, expected):
        """Check that both XML strings represent the same XML tree."""
        checker = LXMLOutputChecker()
        if not checker.check_output(expected, actual, 0):
            message = checker.output_difference(Example("", expected), actual, 0)
            self.fail(message)

    def check_xsd_profile(self, root, sample_file, **substitutions):
        """Check that the SEDA profile can be used to validate a sample XML document."""
        if os.environ.get('TEST_WRITE_SEDA_FILES'):
            fname = join('/tmp', basename(sample_file))
            with open(fname, 'w') as stream:
                stream.write(etree.tostring(root, pretty_print=True))
            print('Generated profile saved to {}'.format(fname))
        profile = self.schema_class(root)
        with open(sample_file) as f:
            sample_xml_string = f.read().format(**substitutions)
        profile.assert_(etree.fromstring(sample_xml_string))


class XMLSchemaTestMixin(XmlTestMixin):
    """Mixin extending XmlTestMixin with implementation of some assert methods for XMLSchema."""

    schema_class = etree.XMLSchema
    schema_file = 'XMLSchema.xsd'
    adapter_id = 'SEDA-2.0.xsd'

    def get_elements(self, profile, name):
        return self.xpath(profile, '//xs:element[@name="{0}"]'.format(name))

    def get_attributes(self, profile, name):
        return self.xpath(profile, '//xs:attribute[@name="{0}"]'.format(name))

    def assertElementDefinition(self, element, expected):
        edef = dict(element.attrib)
        types = self.xpath(element, 'xs:complexType/xs:simpleContent/xs:extension')
        assert len(types) <= 1
        if types:
            edef['type'] = types[0].attrib['base']
        self.assertEqual(edef, expected)

    def assertAttributeDefinition(self, element, expected):
        edef = dict(element.attrib)
        edef.setdefault('use', 'required')
        self.assertEqual(edef, expected)

    def assertXSDAttributes(self, element, expected_attributes):
        # attributes for regular elements
        attrs = [x.attrib for x in self.xpath(element, 'xs:complexType/xs:attribute')]
        # attributes for simple elements
        attrs += [x.attrib for x in self.xpath(
            element, 'xs:complexType/xs:simpleContent/xs:extension/xs:attribute')]
        self.assertEqual(sorted(attrs), expected_attributes)


class RelaxNGTestMixin(XmlTestMixin):
    """Mixin extending XmlTestMixin with implementation of some assert methods for RelaxNG."""

    schema_class = etree.RelaxNG
    schema_file = 'relaxng.rng'
    adapter_id = 'SEDA-2.0.rng'

    def get_elements(self, profile, name):
        return self.xpath(profile, '//rng:element[@name="{0}"]'.format(name))

    def get_attributes(self, profile, name):
        return self.xpath(profile, '//rng:attribute[@name="{0}"]'.format(name))

    def assertElementDefinition(self, element, expected):
        edef = dict(element.attrib)  # {name: element name}
        if element.getparent().tag == '{http://relaxng.org/ns/structure/1.0}optional':
            edef['minOccurs'] = '0'
        elif element.getparent().tag == '{http://relaxng.org/ns/structure/1.0}oneOrMore':
            edef['maxOccurs'] = 'unbounded'
        elif element.getparent().tag == '{http://relaxng.org/ns/structure/1.0}zeroOrMore':
            edef['minOccurs'] = '0'
            edef['maxOccurs'] = 'unbounded'
        refs = self.xpath(element, 'rng:ref')
        if refs:
            assert len(refs) == 1
            edef['type'] = refs[0].attrib['name']
        self._element_fixed_value(edef, element)
        self.assertEqual(edef, expected)

    def assertAttributeDefinition(self, element, expected):
        edef = dict(element.attrib)  # {name: element name}
        if element.getparent().tag == '{http://relaxng.org/ns/structure/1.0}optional':
            edef['use'] = 'optional'
        else:
            edef['use'] = 'required'
        self._element_fixed_value(edef, element)
        self.assertEqual(edef, expected)

    def assertXSDAttributes(self, element, expected_attributes):
        adefs = []
        optattrs = self.xpath(element, 'rng:optional/rng:attribute')
        attrs = self.xpath(element, 'rng:attribute')
        for use, adef_element in chain(izip(repeat('optional'), optattrs),
                                       izip(repeat('required'), attrs)):
            adef = dict(adef_element.attrib)
            adef['use'] = use
            data_elements = self.xpath(adef_element, 'rng:data')
            if data_elements:
                assert len(data_elements) == 1
                adef['type'] = 'xsd:' + data_elements[0].attrib['type']
            self._element_fixed_value(adef, adef_element)
            adefs.append(adef)
        return sorted(adefs)

    def _element_fixed_value(self, edef, element):
        values = self.xpath(element, 'rng:value')
        if values:
            assert len(values) == 1
            value = values[0]
            edef['fixed'] = value.text
            if value.attrib.get('type'):
                edef['type'] = 'xsd:' + value.attrib['type']
        else:
            datatypes = self.xpath(element, 'rng:data')
            if datatypes:
                assert len(datatypes) == 1
                datatype = datatypes[0]
                edef['type'] = 'xsd:' + datatype.attrib['type']


class PathTargetValuesTC(CubicWebTC):

    def test_keyword_path(self):
        element_defs = iter(XSDMMapping('Keyword'))
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity
            kw = create('SEDAKeyword', user_cardinality=u'0..n')
            kt = create('SEDAKeywordType', seda_keyword_type_from=kw)

            edef = next(element_defs)
            # self.assertEqual(
            #  readable_edef(edef),
            #  ('Keyword', 'SEDAKeyword', [
            #      ('id', [('seda_id', 'object', 'SEDAid'),
            #              ('id', 'subject', 'String')]),
            #      ('KeywordContent', []),
            #      ('KeywordReference',
            #       [('seda_keyword_reference_from', 'object', 'SEDAKeywordReference'),
            #        ('seda_keyword_reference_to', 'subject', 'Concept')]),
            #      ('KeywordType', [('seda_keyword_type_from', 'object', 'SEDAKeywordType'),
            #                       ('seda_keyword_type_to', 'subject', 'Concept')]),
            #  ]))
            path = edef[-1][0][1]
            target_values = _path_target_values(kw, path)
            self.assertEqual(len(target_values), 1)
            target_value = target_values[0]
            self.assertEqual(target_value[0], None)
            self.assertEqual(target_value[1], None)

            path = edef[-1][2][1]
            target_values = _path_target_values(kw, path)
            self.assertEqual(len(target_values), 0)

            path = edef[-1][3][1]
            target_values = _path_target_values(kw, path)
            self.assertEqual(len(target_values), 1)
            target_value = target_values[0]
            self.assertEqual(target_value[0].eid, kt.eid)
            self.assertEqual(target_value[1], None)

            kt_scheme = testutils.scheme_for_rtype(cnx, 'seda_keyword_type_to', u'theme')
            kw_type = kt_scheme.reverse_in_scheme[0]
            kt.cw_set(seda_keyword_type_to=kw_type)
            path = edef[-1][3][1]
            target_values = _path_target_values(kw, path)
            self.assertEqual(len(target_values), 1)
            target_value = target_values[0]
            self.assertEqual(target_value[0].eid, kt.eid)
            self.assertEqual(target_value[1].eid, kw_type.eid)

            edef = next(element_defs)
            # self.assertEqual(
            #     readable_edef(edef),
            #     ('KeywordContent', 'SEDAKeyword', [
            #         ('KeywordContent', [('keyword_content', 'subject', 'String')]),
            #     ]))
            path = edef[-1][0][1]
            target_values = _path_target_values(kw, path)
            self.assertEqual(len(target_values), 1)
            target_value = target_values[0]
            self.assertEqual(target_value[0], None)
            self.assertEqual(target_value[1], None)

    def test_internal_reference(self):
        element_defs = iter(XSDMMapping('DataObjectReference'))
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity
            transfer = create('SEDAArchiveTransfer', title=u'test profile')
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(
                transfer)
            bdo = cnx.create_entity('SEDABinaryDataObject',
                                    user_annotation=u'I am mandatory',
                                    seda_binary_data_object=transfer)
            ref = create('SEDADataObjectReference', seda_data_object_reference=unit_alt_seq,
                         seda_data_object_reference_id=bdo)

            edef = next(element_defs)
            # readable_edef(edef)
            # ('DataObjectReference',
            #  'SEDADataObjectReference',
            #  [('id', [('seda_id', 'object', 'SEDAid'), ('id', 'subject', 'String')]),
            #   ('DataObjectReferenceId',
            #    [('seda_data_object_reference_id',
            #      'subject',
            #      ('SEDABinaryDataObject', 'SEDAPhysicalDataObject'))])])

            path = edef[-1][1][1]
            target_values = _path_target_values(ref, path)
            self.assertEqual(len(target_values), 1)
            target_value = target_values[0]
            self.assertEqual(target_value[0], None)
            self.assertEqual(target_value[1].eid, bdo.eid)


class SEDA2RNGExportTC(RelaxNGTestMixin, CubicWebTC):

    def test_skipped_mandatory_simple(self):
        with self.admin_access.cnx() as cnx:
            profile = self.profile_etree(cnx.create_entity('SEDAArchiveTransfer',
                                                           title=u'test profile'))
            date = self.get_element(profile, 'Date')
            self.assertElementDefinition(date, {'name': 'Date',
                                                'type': 'xsd:dateTime'})
            self.assertXSDAttributes(date, [])
            identifier = self.get_element(profile, 'MessageIdentifier')
            self.assertElementDefinition(identifier, {'name': 'MessageIdentifier',
                                                      'type': 'xsd:token'})
            self.assertXSDAttributes(
                identifier,
                [{'name': 'schemeAgencyID', 'use': 'optional', 'type': 'xsd:token'},
                 {'name': 'schemeAgencyName', 'use': 'optional', 'type': 'xsd:string'},
                 {'name': 'schemeDataURI', 'use': 'optional', 'type': 'xsd:anyURI'},
                 {'name': 'schemeID', 'use': 'optional', 'type': 'xsd:token'},
                 {'name': 'schemeName', 'use': 'optional', 'type': 'xsd:string'},
                 {'name': 'schemeURI', 'use': 'optional', 'type': 'xsd:anyURI'},
                 {'name': 'schemeVersionID', 'use': 'optional', 'type': 'xsd:token'}])

    def test_skipped_mandatory_complex(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            testutils.create_data_object(transfer, filename=u'fixed.txt')
            profile = self.profile_etree(transfer)
            fname = self.get_element(profile, 'Filename')
            self.assertElementDefinition(fname, {'name': 'Filename',
                                                 'fixed': 'fixed.txt',
                                                 'type': 'xsd:string'})
            self.assertXSDAttributes(fname, [])
            attachment = self.get_element(profile, 'Attachment')
            self.assertElementDefinition(attachment, {'name': 'Attachment',
                                                      'type': 'xsd:base64Binary'})
            self.assertXSDAttributes(attachment,
                                     [{'name': 'filename', 'use': 'optional',
                                       'type': 'xsd:string'},
                                      {'name': 'uri', 'use': 'optional',
                                       'type': 'xsd:anyURI'}])

    def test_fileinfo_card(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            bdo = cnx.create_entity('SEDABinaryDataObject',
                                    user_annotation=u'I am mandatory',
                                    seda_binary_data_object=transfer)
            appname = cnx.create_entity('SEDACreatingApplicationName',
                                        seda_creating_application_name=bdo)

            profile = self.profile_etree(transfer)
            fileinfo = self.get_element(profile, 'FileInfo')
            self.assertElementDefinition(fileinfo, {'name': 'FileInfo', 'minOccurs': '0'})

            appname.cw_set(user_cardinality=u'1')
            profile = self.profile_etree(transfer)
            fileinfo = self.get_element(profile, 'FileInfo')
            self.assertElementDefinition(fileinfo, {'name': 'FileInfo'})

            appname.cw_set(user_cardinality=u'0..1')
            bdo.cw_set(filename=u'fixed.txt')
            profile = self.profile_etree(transfer)
            fileinfo = self.get_element(profile, 'FileInfo')
            self.assertElementDefinition(fileinfo, {'name': 'FileInfo'})

    def test_data_object_package_card(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            bdo = cnx.create_entity('SEDABinaryDataObject',
                                    user_annotation=u'I am mandatory',
                                    seda_binary_data_object=transfer)

            profile = self.profile_etree(transfer)
            fileinfo = self.get_element(profile, 'DataObjectPackage')
            self.assertElementDefinition(fileinfo, {'name': 'DataObjectPackage', 'minOccurs': '0'})

            bdo.cw_set(user_cardinality=u'1')
            profile = self.profile_etree(transfer)
            dop = self.get_element(profile, 'DataObjectPackage')
            self.assertElementDefinition(dop, {'name': 'DataObjectPackage'})

    def test_object_package_group(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            bdo = cnx.create_entity('SEDABinaryDataObject',
                                    user_annotation=u'I am number one',
                                    seda_binary_data_object=transfer)
            cnx.create_entity('SEDABinaryDataObject',
                              user_annotation=u'I am number two',
                              seda_binary_data_object=transfer)
            cnx.create_entity('SEDAPhysicalDataObject',
                              user_annotation=u'I am number three',
                              seda_physical_data_object=transfer)

            profile = self.profile_etree(transfer)
            dop = self.get_element(profile, 'DataObjectPackage')
            self.assertEqual(len(self.xpath(dop, './rng:group/*')), 3)

            # setting some cardinality to 1 will remove rng:optional parent of the DataObjectPackage
            # and BinaryDataObject nodes
            bdo.cw_set(user_cardinality=u'1')
            profile = self.profile_etree(transfer)
            dop = self.get_element(profile, 'DataObjectPackage')
            self.assertEqual(len(self.xpath(dop, './rng:group/*')), 3)

    def test_transfer_annotation(self):
        with self.admin_access.cnx() as cnx:
            profile = self.profile_etree(cnx.create_entity('SEDAArchiveTransfer',
                                                           title=u'test profile',
                                                           user_annotation=u'some description'))
            docs = self.xpath(profile, '///xs:documentation')
            self.assertEqual(len(docs), 1)
            self.assertEqual(docs[0].text, 'some description')

    def test_transfer_signature(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            cnx.create_entity('SEDASignature', seda_signature=transfer)
            profile = self.profile_etree(transfer)
            signature = self.get_element(profile, 'Signature')
            self.assertElementDefinition(signature, {'name': 'Signature', 'type': 'OpenType',
                                                     'minOccurs': '0'})
            self.assertOpenTypeIsDefined(profile)

    def test_keyword(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity

            scheme = testutils.scheme_for_rtype(cnx, 'seda_keyword_type_to', u'theme')
            kw_type = scheme.reverse_in_scheme[0]

            transfer = create('SEDAArchiveTransfer', title=u'test profile')
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(
                transfer, user_cardinality=u'0..n')

            kw = create('SEDAKeyword', seda_keyword=unit_alt_seq,
                        keyword_content=u'kwick')
            kwr_e = create('SEDAKeywordReference', seda_keyword_reference_from=kw)
            create('SEDAKeywordType', seda_keyword_type_from=kw,
                   seda_keyword_type_to=kw_type)

            profile = self.profile_etree(transfer)

            kwc = self.get_element(profile, 'KeywordContent')
            self.assertElementDefinition(kwc, {'name': 'KeywordContent',
                                               'type': 'xsd:string',
                                               'fixed': 'kwick'})
            kwt = self.get_element(profile, 'KeywordType')
            self.assertElementDefinition(kwt, {'name': 'KeywordType',
                                               'type': 'xsd:token',
                                               'fixed': 'theme',
                                               'minOccurs': '0'})
            self.assertXSDAttributes(
                kwt,
                [{'name': 'listVersionID', 'fixed': 'seda_keyword_type_to/None vocabulary'}])
            kwr = self.get_element(profile, 'KeywordReference')
            self.assertElementDefinition(kwr, {'name': 'KeywordReference',
                                               'type': 'xsd:token'})
            self.assertXSDAttributes(
                kwr,
                [{'name': 'schemeAgencyID', 'type': 'xsd:token', 'use': 'optional'},
                 {'name': 'schemeAgencyName', 'type': 'xsd:string', 'use': 'optional'},
                 {'name': 'schemeDataURI', 'type': 'xsd:anyURI', 'use': 'optional'},
                 {'name': 'schemeID', 'type': 'xsd:token', 'use': 'optional'},
                 {'name': 'schemeName', 'type': 'xsd:string', 'use': 'optional'},
                 {'name': 'schemeURI', 'type': 'xsd:anyURI', 'use': 'optional'},
                 {'name': 'schemeVersionID', 'type': 'xsd:token', 'use': 'optional'}])

            kwr_e.cw_set(seda_keyword_reference_to_scheme=scheme)
            profile = self.profile_etree(transfer)
            kwr = self.get_element(profile, 'KeywordReference')
            self.assertElementDefinition(kwr, {'name': 'KeywordReference',
                                               'type': 'xsd:token'})
            self.assertXSDAttributes(
                kwr,
                [{'name': 'schemeURI', 'fixed': scheme.cwuri}])

            kwr_e.cw_set(seda_keyword_reference_to=kw_type)
            profile = self.profile_etree(transfer)
            kwr = self.get_element(profile, 'KeywordReference')
            self.assertElementDefinition(kwr, {'name': 'KeywordReference',
                                               'type': 'xsd:token',
                                               'fixed': kw_type.cwuri})
            self.assertXSDAttributes(
                kwr,
                [{'name': 'schemeURI', 'fixed': scheme.cwuri}])

    def test_code_list(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            scheme = cnx.create_entity('ConceptScheme', title=u'Keyword Types')
            cnx.create_entity('SEDAMimeTypeCodeListVersion',
                              seda_mime_type_code_list_version_from=transfer,
                              seda_mime_type_code_list_version_to=scheme)

            profile = self.profile_etree(transfer)
            mt_clv = self.get_element(profile, 'MimeTypeCodeListVersion')
            self.assertElementDefinition(mt_clv, {'name': 'MimeTypeCodeListVersion',
                                                  'fixed': scheme.cwuri,
                                                  'type': 'xsd:token',
                                                  'minOccurs': '0'})
            # XXX also fix listSchemeURI ?
            self.assertXSDAttributes(
                mt_clv,
                [{'name': 'listAgencyID', 'use': 'optional', 'type': 'xsd:token'},
                 {'name': 'listAgencyName', 'use': 'optional', 'type': 'xsd:string'},
                 {'name': 'listID', 'use': 'optional', 'type': 'xsd:token'},
                 {'name': 'listName', 'use': 'optional', 'type': 'xsd:string'},
                 {'name': 'listSchemeURI', 'use': 'optional', 'type': 'xsd:anyURI'},
                 {'name': 'listURI', 'use': 'optional', 'type': 'xsd:anyURI'},
                 {'name': 'listVersionID', 'use': 'optional', 'type': 'xsd:token'}])

    def test_seda2_concept(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity
            scheme = create('ConceptScheme', title=u'Digest algorithm')
            some_concept = scheme.add_concept(label=u'md5 algorithm', language_code=u'en')
            transfer = create('SEDAArchiveTransfer', title=u'test profile')
            cnx.create_entity('SEDAMessageDigestAlgorithmCodeListVersion',
                              seda_message_digest_algorithm_code_list_version_from=transfer,
                              seda_message_digest_algorithm_code_list_version_to=scheme)
            create('SEDABinaryDataObject', user_cardinality=u'0..n',
                   user_annotation=u'I am mandatory',
                   seda_binary_data_object=transfer,
                   seda_algorithm=some_concept)

            profile = self.profile_etree(transfer)
            algo = self.get_attribute(profile, 'algorithm')
            self.assertAttributeDefinition(algo, {'name': 'algorithm',
                                                  'use': 'required',
                                                  'type': 'xsd:token',
                                                  'fixed': 'md5 algorithm'})

            create('Label', label_of=some_concept, kind=u'preferred',
                   language_code=u'seda-2', label=u'md5')

            some_concept.cw_clear_all_caches()
            profile = self.profile_etree(transfer)
            algo = self.get_attribute(profile, 'algorithm')
            self.assertAttributeDefinition(algo, {'name': 'algorithm',
                                                  'use': 'required',
                                                  'type': 'xsd:token',
                                                  'fixed': 'md5'})

    def assertOpenTypeIsDefined(self, profile):
        open_types = self.xpath(profile, '//rng:define[@name="OpenType"]')
        self.assertEqual(len(open_types), 1)

    def test_data_duplicates(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(
                transfer, user_cardinality=u'0..n')
            profile = self.profile_etree(transfer)
            title = self.get_element(profile, 'Title')
            self.assertEqual(len(self.xpath(title, 'rng:data')), 1)


class SEDAExportFuncTCMixIn(object):
    """Test that SEDA profile export works correctly."""

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity
            scheme = create('ConceptScheme', title=u'Keyword Types')
            # ensure we're able to export concept with unexpected language code
            some_concept = scheme.add_concept(label=u'md5', language_code=u'de')

            transfer = create('SEDAArchiveTransfer', title=u'test profile')
            cnx.create_entity('SEDAMessageDigestAlgorithmCodeListVersion',
                              seda_message_digest_algorithm_code_list_version_from=transfer,
                              seda_message_digest_algorithm_code_list_version_to=scheme)
            cnx.create_entity('SEDAFileFormatCodeListVersion',
                              seda_file_format_code_list_version_from=transfer)
            create('SEDAMimeTypeCodeListVersion', seda_mime_type_code_list_version_from=transfer,
                   seda_mime_type_code_list_version_to=scheme)
            access_rule = create('SEDAAccessRule', seda_access_rule=transfer)
            access_rule_seq = create('SEDASeqAccessRuleRule',
                                     reverse_seda_seq_access_rule_rule=access_rule)
            create('SEDAStartDate', user_cardinality=u'0..1', seda_start_date=access_rule_seq)
            # binary data object
            bdo = testutils.create_data_object(transfer, user_cardinality=u'0..n',
                                               seda_algorithm=some_concept)
            create('SEDAFormatLitteral', seda_format_litteral=bdo)
            create('SEDAEncoding', seda_encoding_from=bdo)
            create('SEDAUri', seda_uri=bdo.seda_alt_binary_data_object_attachment)
            # first level archive unit
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(
                transfer, user_cardinality=u'0..n', user_annotation=u'Composant ISAD(G)')
            # sub archive unit
            # testutils.create_archive_unit(unit_alt_seq, user_cardinality=u'0..n')
            # management
            appraisal_rule = create('SEDAAppraisalRule', seda_appraisal_rule=unit_alt_seq)
            appraisal_rule_seq = create('SEDASeqAppraisalRuleRule',
                                        reverse_seda_seq_appraisal_rule_rule=appraisal_rule)
            create('SEDAStartDate', user_cardinality=u'0..1', seda_start_date=appraisal_rule_seq)
            access_rule = create('SEDAAccessRule', seda_access_rule=unit_alt_seq)
            create('SEDADisseminationRule', seda_dissemination_rule=unit_alt_seq)
            create('SEDAReuseRule', seda_reuse_rule=unit_alt_seq)
            create('SEDANeedAuthorization', seda_need_authorization=unit_alt_seq)
            # content
            kw = create('SEDAKeyword', user_cardinality=u'0..n', seda_keyword=unit_alt_seq)
            create('SEDAKeywordType', seda_keyword_type_from=kw)
            create('SEDAKeywordReference', seda_keyword_reference_from=kw)
            history_item = create('SEDACustodialHistoryItem',
                                  seda_custodial_history_item=unit_alt_seq)
            create('SEDAwhen', seda_when=history_item)
            version_of = create('SEDAIsVersionOf', seda_is_version_of=unit_alt_seq)
            alt2 = create('SEDAAltIsVersionOfArchiveUnitRefId',
                          reverse_seda_alt_is_version_of_archive_unit_ref_id=version_of)
            create('SEDADataObjectReference', seda_data_object_reference=alt2)
            create('SEDAOriginatingAgencyArchiveUnitIdentifier',
                   seda_originating_agency_archive_unit_identifier=unit_alt_seq)
            create('SEDATransferringAgencyArchiveUnitIdentifier',
                   seda_transferring_agency_archive_unit_identifier=unit_alt_seq)
            create('SEDADescription', seda_description=unit_alt_seq)
            create('SEDALanguage', seda_language_from=unit_alt_seq)
            create('SEDADescriptionLanguage', seda_description_language_from=unit_alt_seq)
            create('SEDACreatedDate', seda_created_date=unit_alt_seq)
            create('SEDAEndDate', seda_end_date=unit_alt_seq)
            create('SEDADataObjectReference', user_cardinality=u'0..n',
                   seda_data_object_reference=unit_alt_seq,
                   seda_data_object_reference_id=bdo)

            cnx.commit()
        self.transfer_eid = transfer.eid
        self.bdo_eid = bdo.eid
        self.au_eid = unit.eid

    def test_profile1(self):
        """Check a minimal SEDA profile validating BV2.0_min.xml."""
        with self.admin_access.cnx() as cnx:
            mda_scheme = cnx.find('ConceptScheme', title=u'Keyword Types').one()
            transfer = cnx.entity_from_eid(self.transfer_eid)
            root = self.profile_etree(transfer)
        self.check_xsd_profile(root, self.datapath('BV2.0_min.xml'),
                               mda_scheme_url=mda_scheme.cwuri)
        # ensure jumped element without content are not there
        self.assertEqual(len(self.get_elements(root, 'Gps')), 0)
        # ensure element with skipped value are not there
        self.assertEqual(len(self.get_elements(root, 'TransactedDate')), 0)
        self.assertProfileDetails(root)


class SEDARNGExportFuncTC(SEDAExportFuncTCMixIn, RelaxNGTestMixin, CubicWebTC):

    def assertProfileDetails(self, root):
        # ensure profile's temporary id are exported in custom seda:profid attribute
        self.assertEqual(len(self.xpath(root, '//rng:attribute[@seda:profid]')), 2)
        for attrdef in self.xpath(root, '//xs:attribute[@seda:profid]'):
            self.assertEqual(attrdef[0]['type'], 'ID')
        # ensure they are properly referenced using 'default' attribute
        xmlid = eid2xmlid(self.bdo_eid)
        references = self.xpath(root, '//rng:element[@a:defaultValue="{}"]'.format(xmlid))
        self.assertEqual(len(references), 1)
        self.assertEqual(references[0].attrib['name'], 'DataObjectReferenceId')
        self.assertEqual(references[0][0].attrib['type'], 'NCName')
        # ensure optional id are properly reinjected
        references = self.xpath(root,
                                '//rng:element[@name="Keyword"]/rng:optional'
                                '/rng:attribute[@name="id"]')
        self.assertEqual(len(references), 1)
        # ensure custodial item content type is properly serialized
        chi = self.xpath(root, '//rng:element[@name="CustodialHistoryItem"]')
        self.assertEqual(len(chi), 1)
        self.assertXSDAttributes(
            chi[0],
            [{'name': 'when', 'use': 'optional', 'type': 'datedateTime'}])
        # ensure types union handling
        ddt = self.xpath(root, '//rng:element[@name="CreatedDate"]/rng:choice/rng:data')
        self.assertEqual(len(ddt), 2)
        self.assertEqual(set(elmt.attrib['type'] for elmt in ddt), set(['date', 'dateTime']))


class OldSEDAExportMixin(object):
    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity

            concepts = {}
            for rtype, etype, value in [
                    ('seda_format_id_to', None, u'fmt/123'),
                    ('seda_encoding_to', None, u'6'),
                    ('seda_type_to', None, u'CDO'),
                    ('seda_description_level', None, u'file'),
                    ('seda_algorithm', 'SEDABinaryDataObject', u'md5'),
                    ('seda_rule', 'SEDASeqAppraisalRuleRule', u'P10Y'),
                    ('seda_rule', 'SEDASeqAccessRuleRule', u'AR038'),
                    ('seda_final_action', 'SEDAAppraisalRule', u'detruire'),
            ]:
                scheme = testutils.scheme_for_type(cnx, rtype, etype, value)
                concepts[value] = scheme.reverse_in_scheme[0]

            # ensure we're able to export concept with unexpected language code
            concepts['md5'].preferred_label[0].cw_set(language_code=u'de')

            agent = testutils.create_authority_record(cnx, u'bob')

            transfer = create('SEDAArchiveTransfer', title=u'my profile title &&',
                              simplified_profile=True)

            create('SEDAComment',
                   user_cardinality=u'1',
                   comment=u'my profile description &&',
                   seda_comment=transfer)

            create('SEDAAccessRule',  # XXX mandatory for seda 1.0
                   user_cardinality=u'1',
                   seda_access_rule=transfer,
                   seda_seq_access_rule_rule=create(
                       'SEDASeqAccessRuleRule', reverse_seda_start_date=create('SEDAStartDate')))
            appraisal_rule_rule = create('SEDASeqAppraisalRuleRule',
                                         seda_rule=concepts['P10Y'],
                                         user_annotation=u"C'est dans 10ans je m'en irai",
                                         reverse_seda_start_date=create('SEDAStartDate'))
            create('SEDAAppraisalRule',
                   seda_appraisal_rule=transfer,
                   seda_final_action=concepts['detruire'],
                   seda_seq_appraisal_rule_rule=appraisal_rule_rule,
                   user_annotation=u'detruire le document')

            _, _, unit_alt_seq = testutils.create_archive_unit(transfer,
                                                               user_cardinality=u'1..n')

            unit_alt_seq.cw_set(reverse_seda_start_date=create('SEDAStartDate',
                                                               user_cardinality=u'0..1'),
                                reverse_seda_end_date=create('SEDAEndDate'),
                                # XXX, value=date(2015, 2, 24)),
                                reverse_seda_description=create('SEDADescription'))

            kw = create('SEDAKeyword',
                        user_cardinality=u'0..n',
                        seda_keyword=unit_alt_seq)
            create('SEDAKeywordReference',
                   seda_keyword_reference_from=kw,
                   seda_keyword_reference_to=concepts['file'],
                   seda_keyword_reference_to_scheme=concepts['file'].scheme)
            create('SEDAKeywordType',
                   seda_keyword_type_from=kw,
                   user_cardinality=u'0..1')

            create('SEDAOriginatingAgency', seda_originating_agency_from=unit_alt_seq,
                   seda_originating_agency_to=agent)

            create('SEDAType', seda_type_from=unit_alt_seq,
                   seda_type_to=concepts['CDO'])

            create('SEDACustodialHistoryItem', seda_custodial_history_item=unit_alt_seq,
                   reverse_seda_when=create('SEDAwhen'))

            # Add sub archive unit
            _, _, subunit_alt_seq = testutils.create_archive_unit(unit_alt_seq,
                                                                  user_cardinality=u'1..n')

            create('SEDAAppraisalRule',
                   seda_appraisal_rule=subunit_alt_seq,
                   seda_seq_appraisal_rule_rule=create(
                       'SEDASeqAppraisalRuleRule', reverse_seda_start_date=create('SEDAStartDate')))

            create('SEDAAccessRule',
                   user_cardinality=u'1',
                   user_annotation=u'restrict',
                   seda_access_rule=subunit_alt_seq,
                   seda_seq_access_rule_rule=create('SEDASeqAccessRuleRule',
                                                    reverse_seda_start_date=create('SEDAStartDate'),
                                                    seda_rule=concepts['AR038']))

            # Add minimal document to first level archive
            ref = create('SEDADataObjectReference', seda_data_object_reference=unit_alt_seq)
            bdo = testutils.create_data_object(transfer, user_cardinality=u'0..n',
                                               filename=u'this_is_the_filename.pdf',
                                               seda_algorithm=concepts['md5'],
                                               reverse_seda_data_object_reference_id=ref)

            create('SEDAFormatId',
                   user_cardinality=u'1',
                   seda_format_id_from=bdo,
                   seda_format_id_to=concepts['fmt/123'])
            create('SEDAEncoding',
                   user_cardinality=u'1',
                   seda_encoding_from=bdo,
                   seda_encoding_to=concepts['6'])

            # Add another sub archive unit
            _, _, subunit2_alt_seq = testutils.create_archive_unit(unit_alt_seq,
                                                                   user_cardinality=u'1..n')

            cnx.commit()

        self.transfer_eid = transfer.eid
        self.bdo_eid = bdo.eid
        self.file_concept_eid = concepts['file'].eid
        self.agent_eid = agent.eid

    def _test_profile(self, adapter_id, expected_file):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            file_concept = cnx.entity_from_eid(self.file_concept_eid)
            agent = cnx.entity_from_eid(self.agent_eid)

            adapter = transfer.cw_adapt_to(adapter_id)
            generated_xsd = adapter.dump()

            if os.environ.get('TEST_WRITE_SEDA_FILES'):
                orig_content = generated_xsd
                for value, key in [(file_concept.cwuri, 'concept-uri'),
                                   (file_concept.scheme.cwuri, 'scheme-uri'),
                                   (text_type(agent.eid), 'agent-id'),
                                   (agent.dc_title(), 'agent-name')]:
                    orig_content = orig_content.replace(value, '%({})s'.format(key))
                with open(self.datapath(expected_file + '.new'), 'w') as stream:
                    stream.write(orig_content)
                print('Regenerated expected file as {}.new'.format(expected_file))

            root = etree.fromstring(generated_xsd)
            self.assertXmlValid(root)
            with open(self.datapath(expected_file)) as expected:
                self.assertXmlEqual(expected.read()
                                    % {'bdo-eid': binary_type(self.bdo_eid),
                                       'concept-uri': binary_type(file_concept.cwuri),
                                       'scheme-uri': binary_type(file_concept.scheme.cwuri),
                                       'agent-id': binary_type(agent.eid),
                                       'agent-name': binary_type(agent.dc_title())},
                                    generated_xsd)
            return adapter, root


class OldSEDAXSDExportTC(XMLSchemaTestMixin, OldSEDAExportMixin, CubicWebTC):

    def test_seda_1_0(self):
        self._test_profile('SEDA-1.0.xsd', 'seda_1_export.xsd')

    def test_seda_0_2(self):
        self._test_profile('SEDA-0.2.xsd', 'seda_02_export.xsd')

    def _test_profile(self, adapter_id, expected_file):
        adapter, root = super(OldSEDAXSDExportTC, self)._test_profile(adapter_id, expected_file)
        # ensure there is no element with @type but a complex type
        namespaces = adapter.namespaces.copy()
        namespaces.pop(None)
        dates = root.xpath('//xsd:element[@name="Date"]/xsd:complexType/xsd:sequence',
                           namespaces=namespaces)
        self.assertEqual(len(dates), 0)


class OldSEDARNGExportTC(RelaxNGTestMixin, OldSEDAExportMixin, CubicWebTC):

    def test_seda_1_0(self):
        self._test_profile('SEDA-1.0.rng', 'seda_1_export.rng')

    def test_seda_0_2_rng(self):
        self._test_profile('SEDA-0.2.rng', 'seda_02_export.rng')

    def test_seda_0_2_bordereau_ref(self):
        """Check a sample SEDA 0.2 profile validation."""
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity

            transfer = create('SEDAArchiveTransfer', title=u'test profile',
                              simplified_profile=True)
            create('SEDAComment', seda_comment=transfer)

            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            create('SEDAArchivalAgreement', seda_archival_agreement=transfer)
            create('SEDATransferringAgencyArchiveUnitIdentifier',
                   seda_transferring_agency_archive_unit_identifier=unit_alt_seq)
            create('SEDAStartDate', seda_start_date=unit_alt_seq)
            create('SEDAEndDate', seda_end_date=unit_alt_seq)
            appraisal_rule = create('SEDAAppraisalRule', seda_appraisal_rule=unit_alt_seq)
            appraisal_rule_seq = create('SEDASeqAppraisalRuleRule',
                                        reverse_seda_seq_appraisal_rule_rule=appraisal_rule)
            create('SEDAStartDate', seda_start_date=appraisal_rule_seq)
            access_rule = create('SEDAAccessRule', seda_access_rule=unit_alt_seq)
            access_rule_seq = create('SEDASeqAccessRuleRule',
                                     reverse_seda_seq_access_rule_rule=access_rule)
            create('SEDAStartDate', seda_start_date=access_rule_seq)

            subunit, subunit_alt, subunit_alt_seq = testutils.create_archive_unit(
                unit_alt_seq)
            create('SEDATransferringAgencyArchiveUnitIdentifier',
                   seda_transferring_agency_archive_unit_identifier=subunit_alt_seq)
            create('SEDAStartDate', seda_start_date=subunit_alt_seq)
            create('SEDAEndDate', seda_end_date=subunit_alt_seq)
            kw = create('SEDAKeyword', user_cardinality=u'0..n', seda_keyword=subunit_alt_seq)
            create('SEDAKeywordReference', seda_keyword_reference_from=kw)

            create('SEDASystemId', seda_system_id=subunit_alt_seq)

            bdo = testutils.create_data_object(transfer)
            create('SEDADataObjectReference',
                   seda_data_object_reference=subunit_alt_seq,
                   seda_data_object_reference_id=bdo)
            create('SEDAEncoding', seda_encoding_from=bdo)
            create('SEDAMimeType', seda_mime_type_from=bdo)
            create('SEDADateCreatedByApplication', seda_date_created_by_application=bdo)

            cnx.commit()

            root = self.profile_etree(transfer, 'SEDA-0.2.rng')
        self.check_xsd_profile(root, self.datapath('seda_02_bordereau_ref.xml'))


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda unit tests for XSD profile generation"""

from collections import namedtuple

from lxml import etree

from cubicweb.devtools.testlib import CubicWebTC

from test_profile_generation import XmlTestMixin, SEDAExportFuncTCMixIn


AttrDef = namedtuple('AttrDef', ['label', 'card', 'value'])


class SEDAHTMLExportFuncTC(SEDAExportFuncTCMixIn, XmlTestMixin, CubicWebTC):
    adapter_id = 'SEDA-2.0.html'

    def setup_database(self):
        super(SEDAHTMLExportFuncTC, self).setup_database()
        with self.admin_access.cnx() as cnx:
            scheme = cnx.find('ConceptScheme', title=u'Keyword Types').one()
            some_concept = scheme.reverse_in_scheme[0]
            lang_rtype = cnx.find('CWRType', name='seda_language_to').one()
            some_concept.scheme.cw_set(scheme_relation_type=lang_rtype)
            cnx.find('SEDALanguage').one().cw_set(seda_language_to=some_concept)
            cnx.find('SEDAKeywordReference').one().cw_set(
                seda_keyword_reference_to=some_concept,
                seda_keyword_reference_to_scheme=some_concept.scheme)
            cnx.commit()

        self.concept_eid = some_concept.eid

    def assertXmlValid(self, root):
        pass  # no schema available for html 5

    def get_element(self, profile, name):
        elements = self.xpath(profile, '//h3[text()="{0}"]'.format(name))
        self.assertEqual(len(elements), 1)
        return elements[0].getparent()

    def element_values(self, element):
        el_defs = []
        for div in self.xpath(element, 'div'):
            el_def = {}
            el_defs.append(el_def)
            for span in self.xpath(div, 'span'):
                if span:
                    value = etree.tostring(span[0])
                else:
                    value = span.text
                el_def[span.attrib['class']] = value
        return el_defs

    def test_profile1(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            some_concept = cnx.entity_from_eid(self.concept_eid)
            profile = self.profile_etree(transfer)
            # ensure encoding is declared
            meta = self.xpath(profile, '/html/head/meta')[0]
            self.assertEqual(meta.attrib, {'content': 'text/html; charset=UTF-8',
                                           'http-equiv': 'content-type'})
            # ensure only one first level title
            self.assertEqual(len(self.xpath(profile, '//title')), 1)
            self.assertEqual(len(self.xpath(profile, '//h1')), 1)
            # ensure annotation are serialized
            self.assertEqual(
                [ann for ann in self.xpath(profile, '//div[@class="description"]/text()')
                 if ann != 'data object title'],
                ['Composant ISAD(G)'])
            # ensure all attributes have label, card and value defined
            attr_divs = self.xpath(profile, '//div[@class="attribute"]')
            attr_defs = set()
            for attr_div in attr_divs:
                attr_def = AttrDef(self.xpath(attr_div, 'span[@class="label"]')[0].text,
                                   self.xpath(attr_div, 'span[@class="card"]')[0].text,
                                   self.xpath(attr_div, 'span[@class="value"]')[0].text)
                attr_defs.add(attr_def)
            self.assertEqual(attr_defs, set([
                AttrDef(label='algorithm', card='mandatory', value=None),
                AttrDef(label='filename', card='optional', value=None),
                AttrDef(label='id', card='mandatory', value='id{}'.format(self.au_eid)),
                AttrDef(label='id', card='mandatory', value='id{}'.format(self.bdo_eid)),
                AttrDef(label='id', card='optional', value=None),
                AttrDef(label='when', card='optional', value=None),
                AttrDef(label='uri', card='optional', value=None),
            ]))
            # ensure jumped element children have proper parent
            clv = self.get_element(profile, 'CodeListVersions')
            self.assertEqual(len(self.xpath(clv, 'div/h3[text()="MimeTypeCodeListVersion"]')), 1)
            # ensure KeywordType list handling
            ktype = self.get_element(profile, 'KeywordType')
            el_defs = self.element_values(ktype)
            self.assertEqual(el_defs, [{'label': 'listVersionID', 'value': 'edition 2009'},
                                       {'label': 'XSD content type', 'value': 'xsd:token'}])
            # KeywordReference
            kref = self.get_element(profile, 'KeywordReference')
            el_defs = self.element_values(kref)
            concept_link = '<a href="{0}">md5</a>'.format(some_concept.absolute_url())
            scheme = some_concept.scheme
            scheme_link = '<a href="{0}">Keyword Types</a>'.format(scheme.absolute_url())
            self.assertEqual(el_defs, [{'label': 'schemeURI', 'value': scheme_link},
                                       {'label': 'XSD content type', 'value': 'xsd:token'},
                                       {'label': 'fixed value', 'value': concept_link}])
            # Language
            lang = self.get_element(profile, 'Language')
            el_defs = self.element_values(lang)
            self.assertEqual(el_defs, [{'label': 'XSD content type', 'value': 'xsd:language'},
                                       {'label': 'fixed value', 'value': concept_link}])


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2016-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda unit tests for entities.container"""

import json
import unittest

from six import text_type

from logilab.common import attrdict

from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_seda.entities import (seda_profile_container_def, simplified_profile,
                                    full_seda2_profile, parent_and_container,
                                    rule_type_from_etype, custom)

from cubicweb_seda.testutils import create_archive_unit, create_data_object


def sort_container(container_def):
    for k, v in container_def:
        yield k, sorted(v)


class ContainerTC(CubicWebTC):

    def test_seda_profile_container(self):
        # line below should be copied from entities.container.registration_callback
        container_def = seda_profile_container_def(self.schema)
        container_def = dict(sort_container(container_def))
        self.assertEqual(container_def['SEDAMimeType'],
                         [('seda_mime_type_from', 'subject')])
        self.assertNotIn('ConceptScheme', container_def)
        self.assertNotIn('Concept', container_def)
        self.assertNotIn('AuthorityRecord', container_def)
        entity = self.vreg['etypes'].etype_class('SEDAArchiveTransfer')(self)
        self.assertIsNotNone(entity.cw_adapt_to('IContainer'))
        self.assertIsNone(entity.cw_adapt_to('IContained'))

    def test_container_relation(self):
        with self.admin_access.cnx() as cnx:
            create = cnx.create_entity
            transfer = create('SEDAArchiveTransfer', title=u'test profile')
            mtclv = create('SEDAMimeTypeCodeListVersion',
                           seda_mime_type_code_list_version_from=transfer)
            access_rule = create('SEDAAccessRule', seda_access_rule=transfer)
            cnx.commit()
            for entity in (mtclv, access_rule):
                entity.cw_clear_all_caches()
                self.assertEqual(entity.cw_adapt_to('IContained').container.eid, transfer.eid)
            access_rule_seq = create('SEDASeqAccessRuleRule',
                                     reverse_seda_seq_access_rule_rule=access_rule)
            start_date = create('SEDAStartDate', seda_start_date=access_rule_seq)
            cnx.commit()
            for entity in (access_rule_seq, start_date):
                entity.cw_clear_all_caches()
                self.assertEqual(entity.cw_adapt_to('IContained').container.eid, transfer.eid)

    def test_archive_unit_container_clone(self):
        """Functional test for SEDA component clone."""
        with self.admin_access.cnx() as cnx:
            unit, unit_alt, unit_alt_seq = create_archive_unit(None, cnx=cnx)
            bdo = create_data_object(unit_alt_seq)
            cnx.commit()

            unit.cw_clear_all_caches()
            self.assertEqual(unit.container, ())  # XXX arguable
            unit_alt_seq.cw_clear_all_caches()
            self.assertEqual(unit_alt_seq.container[0].eid, unit.eid)
            bdo.cw_clear_all_caches()
            self.assertEqual(bdo.container[0].eid, unit.eid)

            # test clone without reparenting
            cloned = cnx.create_entity(unit.cw_etype, user_annotation=u'clone',
                                       clone_of=unit)
            cnx.commit()
            cloned.cw_clear_all_caches()
            self.assertEqual(cloned.user_annotation, 'clone')
            cloned_unit_alt_seq = cloned.first_level_choice.content_sequence
            self.assertEqual(cloned_unit_alt_seq.container[0].eid, cloned.eid)
            cloned_bdo = (cloned_unit_alt_seq.reverse_seda_data_object_reference[0]
                          .seda_data_object_reference_id[0])
            self.assertEqual(cloned_bdo.container[0].eid, cloned.eid)

            # test clone with reparenting
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            cloned = cnx.create_entity(unit.cw_etype,
                                       user_annotation=u'I am mandatory',
                                       clone_of=unit,
                                       seda_archive_unit=transfer)
            cnx.commit()
            cloned.cw_clear_all_caches()
            self.assertEqual(cloned.container[0].eid, transfer.eid)
            cloned.cw_clear_all_caches()
            cloned_unit_alt_seq = cloned.first_level_choice.content_sequence
            self.assertEqual(cloned_unit_alt_seq.container[0].eid, transfer.eid)
            cloned_bdo = (cloned_unit_alt_seq.reverse_seda_data_object_reference[0]
                          .seda_data_object_reference_id[0])
            self.assertEqual(cloned_bdo.container[0].eid, transfer.eid)
            cloned_bdo.cw_clear_all_caches()
            self.assertEqual(cloned_bdo.seda_binary_data_object[0].eid, transfer.eid)

    def test_archive_transfer_clone(self):
        """Functional test for SEDA profile cloning."""
        with self.admin_access.cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'Algorithms')
            concept = scheme.add_concept(u'md5')
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            cnx.create_entity('SEDAMessageDigestAlgorithmCodeListVersion',
                              seda_message_digest_algorithm_code_list_version_from=transfer,
                              seda_message_digest_algorithm_code_list_version_to=scheme)
            unit, unit_alt, unit_alt_seq = create_archive_unit(
                transfer, cnx=cnx, title=u'hello')
            create_data_object(transfer, seda_algorithm=concept)
            cnx.commit()

            clone = cnx.create_entity('SEDAArchiveTransfer', title=u'Clone', clone_of=transfer)
            cnx.commit()

            self.assertEqual((clone.reverse_seda_message_digest_algorithm_code_list_version_from[0].
                             seda_message_digest_algorithm_code_list_version_to[0].eid),
                             scheme.eid)
            seq = clone.archive_units[0].first_level_choice.content_sequence
            self.assertEqual(seq.title.title, 'hello')
            self.assertEqual(transfer.binary_data_objects[0].seda_algorithm[0].eid, concept.eid)

    def test_container_clone_clone(self):
        """Functional test for SEDA component clone."""
        with self.admin_access.cnx() as cnx:
            unit, unit_alt, unit_alt_seq = create_archive_unit(None, cnx=cnx)
            cnx.commit()

            cloned = cnx.create_entity(unit.cw_etype, user_annotation=u'x', clone_of=unit)
            cnx.commit()
            cloned.cw_clear_all_caches()

            cloned2 = cnx.create_entity(unit.cw_etype, user_annotation=u'Y')
            # mimick what's occurs when you're using the copy form
            cloned2.copy_relations(cloned.eid)  # 1. copy relation
            self.assertFalse(cloned2.clone_of)
            cloned2.cw_set(clone_of=cloned.eid)  # 2. execute __linkto (which trigger cloning)

            cnx.commit()

            cloned2.cw_clear_all_caches()
            self.assertEqual(len(cloned2.seda_alt_archive_unit_archive_unit_ref_id), 1)

            unit_alt.cw_clear_all_caches()
            self.assertEqual(unit_alt.container[0].eid, unit.eid)
            self.assertEqual(cloned.seda_alt_archive_unit_archive_unit_ref_id[0].container[0].eid,
                             cloned.eid)
            self.assertEqual(cloned2.seda_alt_archive_unit_archive_unit_ref_id[0].container[0].eid,
                             cloned2.eid)


class FakeEntity(object):
    cw_etype = 'Whatever'

    def __init__(self, _cw):
        self._cw = _cw

    def has_eid(self):
        return False


class PredicatesTC(CubicWebTC):

    def test_simplified_profile(self):
        simplified_profile_pred = simplified_profile()
        full_seda2_profile_pred = full_seda2_profile()
        with self.admin_access.web_request() as req:
            transfer = req.create_entity('SEDAArchiveTransfer', title=u'test profile')
            self.assertEqual(simplified_profile_pred(None, req, entity=transfer), 0)
            self.assertEqual(full_seda2_profile_pred(None, req, entity=transfer), 1)
            self.assertEqual(simplified_profile_pred(None, req, rset=transfer.as_rset()), 0)
            self.assertEqual(full_seda2_profile_pred(None, req, rset=transfer.as_rset()), 1)
            access_rule = req.create_entity('SEDAAccessRule', seda_access_rule=transfer)
            req.cnx.commit()  # needed to set the container relation
            self.assertEqual(simplified_profile_pred(None, req, entity=access_rule), 0)
            self.assertEqual(full_seda2_profile_pred(None, req, entity=access_rule), 1)
            req.form['__linkto'] = 'whatever:%s:whatever' % transfer.eid
            being_created = FakeEntity(req)
            self.assertEqual(simplified_profile_pred(None, req, entity=being_created), 0)
            self.assertEqual(full_seda2_profile_pred(None, req, entity=being_created), 1)
            del req.form['__linkto']
            req.form['arg'] = [text_type(transfer.eid)]
            self.assertEqual(simplified_profile_pred(None, req, entity=being_created), 0)
            self.assertEqual(full_seda2_profile_pred(None, req, entity=being_created), 1)
            transfer.cw_set(simplified_profile=True)
            self.assertEqual(simplified_profile_pred(None, req, entity=transfer), 1)
            self.assertEqual(full_seda2_profile_pred(None, req, entity=transfer), 0)

            req.form = {'etype': 'SEDAArchiveUnit'}
            etype_vreg = req.vreg['etypes']
            unit = etype_vreg.etype_class('SEDAArchiveUnit')(req)
            self.assertEqual(simplified_profile_pred(None, req, entity=unit), 1)
            entity = etype_vreg.etype_class('SEDASeqAltArchiveUnitArchiveUnitRefIdManagement')(req)
            self.assertEqual(simplified_profile_pred(None, req, entity=entity), 1)


class ITreeTC(CubicWebTC):

    def assertChildren(self, entity, expected_eids):
        itree = entity.cw_adapt_to('ITreeBase')
        children = [x.eid for x in itree.iterchildren()]
        self.assertEqual(children, expected_eids)

    def assertParent(self, entity, expected_eid):
        itree = entity.cw_adapt_to('ITreeBase')
        parent = itree.parent()
        if parent:
            parent_eid = parent.eid
        else:
            parent_eid = None
        self.assertEqual(parent_eid, expected_eid)

    def test(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            au, alt, seq = create_archive_unit(transfer)
            do_ref = cnx.create_entity('SEDADataObjectReference',
                                       seda_data_object_reference=seq)
            bdo = create_data_object(transfer,
                                     reverse_seda_data_object_reference_id=do_ref)
            cnx.commit()
            au.cw_clear_all_caches()
            bdo.cw_clear_all_caches()

            self.assertChildren(transfer, [bdo.eid, au.eid])
            self.assertChildren(au, [])
            self.assertParent(transfer, None)
            self.assertParent(au, transfer.eid)
            self.assertParent(bdo, transfer.eid)

            transfer.cw_set(simplified_profile=True)
            cnx.commit()
            au.cw_clear_all_caches()
            bdo.cw_clear_all_caches()

            self.assertChildren(transfer, [au.eid])
            self.assertChildren(au, [bdo.eid])
            self.assertParent(transfer, None)
            self.assertParent(au, transfer.eid)
            self.assertParent(bdo, au.eid)


class RuleFromETypeTC(unittest.TestCase):
    def test_rule_from_etype(self):
        for rule_type in ('access', 'appraisal', 'classification',
                          'reuse', 'dissemination', 'storage'):
            for prefix, suffix in [
                    ('SEDAAlt', 'RulePreventInheritance'),
                    ('SEDASeq', 'RuleRule'),
                    ('SEDA', 'Rule'),
            ]:
                self.assertEqual(rule_type_from_etype(prefix + rule_type.capitalize() + suffix),
                                 rule_type)


class ParentAndContainerTC(CubicWebTC):

    def test_nodata(self):
        with self.admin_access.web_request() as req:
            parent, container = parent_and_container(attrdict(_cw=req, has_eid=lambda: False))
            self.assertIsNone(parent)
            self.assertIsNone(container)

    def test_linkto(self):
        with self.admin_access.web_request() as req:
            transfer = req.cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            req.form['__linkto'] = 'x:{0}:y'.format(transfer.eid)
            parent, container = parent_and_container(attrdict(_cw=req, has_eid=lambda: False))
            self.assertEqual(parent.eid, transfer.eid)
            self.assertEqual(container.eid, transfer.eid)

    def test_arg(self):
        with self.admin_access.web_request() as req:
            transfer = req.cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            req.form['arg'] = [json.dumps(transfer.eid)]
            parent, container = parent_and_container(attrdict(_cw=req, has_eid=lambda: False))
            self.assertEqual(parent.eid, transfer.eid)
            self.assertEqual(container.eid, transfer.eid)

    def test_eid(self):
        with self.admin_access.web_request() as req:
            transfer = req.cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            req.form['sedaContainerEID'] = text_type(transfer.eid)
            parent, container = parent_and_container(attrdict(_cw=req, has_eid=lambda: False))
            self.assertIsNone(parent)
            self.assertEqual(container.eid, transfer.eid)


class CustomEntitiesTC(CubicWebTC):

    def test_title(self):
        with self.admin_access.cnx() as cnx:
            for etype in ('SEDAArchiveUnit', 'SEDABinaryDataObject', 'SEDAPhysicalDataObject'):
                ent = cnx.create_entity(etype, user_annotation=u'bla bla\nbli bli blo\n')
                self.assertEqual(ent.dc_title(), u'bla bla')

    def test_climb_rule_holders(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.create_entity('SEDAArchiveTransfer', title=u'test profile')
            unit, _, unit_alt_seq = create_archive_unit(transfer)
            subunit, _, subunit_alt_seq = create_archive_unit(unit_alt_seq)
            cnx.commit()

            self.assertEqual(list(custom._climb_rule_holders(subunit_alt_seq)),
                             [subunit_alt_seq, unit_alt_seq, transfer])
            self.assertEqual(list(custom._climb_rule_holders(subunit)),
                             [subunit_alt_seq, unit_alt_seq, transfer])
            self.assertEqual(list(custom._climb_rule_holders(unit_alt_seq)),
                             [unit_alt_seq, transfer])
            self.assertEqual(list(custom._climb_rule_holders(unit)),
                             [unit_alt_seq, transfer])
            self.assertEqual(list(custom._climb_rule_holders(transfer)),
                             [transfer])


if __name__ == '__main__':
    unittest.main()



# copyright 2016-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Custom logic (eg entities and adapters) for cubicweb-seda."""

import json

from logilab.common.registry import objectify_predicate

from cubicweb.predicates import is_instance
from cubicweb_compound.entities import IContainer, IContained, IClonableAdapter

from .. import seda_profile_container_def
from ..xsd import XSDMMapping
from ..xsd2yams import RULE_TYPES


XSDM_MAPPING = XSDMMapping('ArchiveTransfer')


def parent_and_container(entity):
    """Attempt to return the direct parent and container from the entity, handling case where entity
    is being created and container information will be found through linkto or parent form
    (partially supported)
    """
    if entity.has_eid():
        # entity is expected to be a contained entity, not the container itself
        container = entity.cw_adapt_to('IContained').container
        parent = entity.cw_adapt_to('IContained').parent
        if container is None:
            # entity may be both container and contained, and in this case is a container
            assert entity.cw_adapt_to('IContainer'), entity
            container = entity
    else:
        req = entity._cw
        # but parent entity, retrieved through linkto, may be the container itself or a
        # contained entity
        try:
            parent_eid = int(req.form['__linkto'].split(':')[1])
        except KeyError:
            # ajax created form
            try:
                parent_eid = int(json.loads(req.form['arg'][0]))
            except (KeyError, ValueError):
                if 'sedaContainerEID' in req.form:
                    container = req.entity_from_eid(int(req.form['sedaContainerEID']))
                    return None, container
                if 'referenced_by' in req.form:
                    entity = req.entity_from_eid(int(req.form['referenced_by']))
                    container = entity.cw_adapt_to('IContained').container
                    return None, container
                # unable to get parent eid for now :(
                return None, None
        parent = req.entity_from_eid(parent_eid)
        # handle IContained first: in case entity support both interface we want to go to the
        # uppermost parent
        icontained = parent.cw_adapt_to('IContained')
        if icontained is not None and icontained.container:
            container = icontained.container
        elif parent.cw_adapt_to('IContainer'):
            container = parent
    return parent, container


def _seda_container_from_context(rset, entity, **kwargs):
    if entity is None:
        if 'row' in kwargs:
            entity = rset.get_entity(kwargs['row'], kwargs.get('col', 0))
        else:
            entity = rset.one()
    # protect against unrelated entity types
    if not entity.cw_etype.startswith('SEDA'):
        return None
    if entity.cw_etype != 'SEDAArchiveTransfer':
        entity = parent_and_container(entity)[1]
    return entity


@objectify_predicate
def component_unit(cls, req, rset=None, entity=None, **kwargs):
    """Predicate returning 1 score if context entity is within "component" archive unit (i.e.
    container root is not a SEDAArchiveTransfer but a SEDAArchiveUnit).
    """
    entity = _seda_container_from_context(rset, entity, **kwargs)
    return 1 if entity.cw_etype == 'SEDAArchiveUnit' else 0


@objectify_predicate
def simplified_profile(cls, req, rset=None, entity=None, **kwargs):
    """Predicate returning 1 score if context entity is within a simplified profile."""
    container = _seda_container_from_context(rset, entity, **kwargs)
    if container is None:
        # Detect creation of a component archive unit, which are supposed to be
        # simplified. If we can't get a container and we're creating an archive
        # unit, it must be an archive unit component (else it's parent container
        # should have been retrieved)
        if getattr(req, 'form', {}).get('etype') == 'SEDAArchiveUnit':
            return 1
        return 0
    if container.cw_etype == 'SEDAArchiveUnit':
        # XXX archive unit component, for now suppose it's "simplified"
        return 1
    return 1 if container.simplified_profile else 0


def is_full_seda2_profile(entity=None, rset=None, **kwargs):
    """Return 1 if context entity is within a full seda2 profile, else 0."""
    entity = _seda_container_from_context(rset, entity, **kwargs)
    if entity is None:
        return 1
    if entity.cw_etype == 'SEDAArchiveUnit':
        # XXX archive unit component, for now suppose it's "simplified"
        return 0
    return 0 if entity.simplified_profile else 1


@objectify_predicate
def full_seda2_profile(cls, req, rset=None, entity=None, **kwargs):
    """Predicate returning 1 score if context entity is within a full seda2 profile."""
    return is_full_seda2_profile(entity, rset, **kwargs)


def rule_type_from_etype(etype):
    """Return the rule type (e.g. 'access') from an etype enclosing the information
    (e.g. 'SEDAAltAccessRulePreventInheritance', 'SEDASeqAaccessRuleRule' or 'SEDAAccessRule')
    """
    if etype.startswith('SEDAAlt'):
        rule_type = etype[len('SEDAAlt'):-len('RulePreventInheritance')]
    elif etype.startswith('SEDASeq'):
        rule_type = etype[len('SEDASeq'):-len('RuleRule')]
    else:
        rule_type = etype[len('SEDA'):-len('Rule')]
    rule_type = rule_type.lower()
    assert rule_type in RULE_TYPES, 'unhandled etype {0}'.format(etype)
    return rule_type


class DirectLinkIContained(IContained):
    """IContained implementation using a relation that link every contained entities to its parent
    container.
    """
    __abstract__ = True

    @property
    def container(self):
        """Return the container to which this entity belongs, or None."""
        container = self.entity.related('container', entities=True)
        return container and container[0] or None


class SEDAArchiveTransferIClonableAdapter(IClonableAdapter):
    """Cloning adapter for SEDA profiles."""
    __select__ = is_instance('SEDAArchiveTransfer')
    rtype = 'clone_of'
    skiprtypes = ('container',)


class SEDAArchiveUnitIClonableAdapter(IClonableAdapter):
    """Cloning adapter for SEDA components."""
    __select__ = is_instance('SEDAArchiveUnit')
    rtype = 'clone_of'
    skiprtypes = ('container',)

    def clone_into(self, clone):
        """Recursivily clone the container graph of this entity into `clone`."""
        if clone.seda_archive_unit and (
                clone.seda_archive_unit[0].cw_etype == 'SEDAArchiveTransfer'
                or clone.seda_archive_unit[0].container[0].cw_etype == 'SEDAArchiveTransfer'):
            # clone is parented to a transfer profile, we need to properly handle binary/physical
            # data objects
            data_objects = self._cw.execute(
                'Any X WHERE X is IN (SEDABinaryDataObject, SEDAPhysicalDataObject),'
                ' X container %(c)s', {'c': self.entity.eid})
        else:
            data_objects = None
        clones = super(SEDAArchiveUnitIClonableAdapter, self).clone_into(clone)
        if data_objects is not None:
            if clone.seda_archive_unit[0].cw_etype == 'SEDAArchiveTransfer':
                transfer = clone.seda_archive_unit[0]
            else:
                transfer = clone.seda_archive_unit[0].container[0]
            for data_object in data_objects.entities():
                rtype = {
                    'SEDABinaryDataObject': 'seda_binary_data_object',
                    'SEDAPhysicalDataObject': 'seda_physical_data_object',
                }[data_object.cw_etype]
                clones[data_object].cw_set(**{rtype: transfer})


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__)
    vreg.register(IContainer.build_class('SEDAArchiveTransfer'))
    vreg.register(IContainer.build_class('SEDAArchiveUnit'))  # archive unit may also be a container
    for etype, parent_relations in sorted(seda_profile_container_def(vreg.schema)):
        cls = DirectLinkIContained.build_class(etype, parent_relations)
        assert cls
        vreg.register(cls)
        if etype in ('SEDABinaryDataObject', 'SEDAPhysicalDataObject'):
            # control parent relation order
            cls.parent_relations = list(cls.parent_relations)
            if cls.parent_relations[0][0] == 'seda_data_object_reference_id':
                cls.parent_relations = cls.parent_relations[::-1]



# copyright 2016-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from ..xsd2yams import RULE_TYPES
from . import generated, itree


def _extract_title(annotation):
    """Return the first line in the annotation to use as a title"""
    annotation = annotation.strip()
    assert annotation
    return annotation.splitlines()[0]


def _climb_rule_holders(transfer_or_archive_unit):
    """Starting from a transfer or archive unit entity, yield entity that may be linked to
    management rule until the root (transfer) is reached.
    """
    while transfer_or_archive_unit is not None:
        if transfer_or_archive_unit.cw_etype == 'SEDASeqAltArchiveUnitArchiveUnitRefIdManagement':
            yield transfer_or_archive_unit
            transfer_or_archive_unit = itree.parent_archive_unit(transfer_or_archive_unit)
        elif transfer_or_archive_unit.cw_etype == 'SEDAArchiveTransfer':
            yield transfer_or_archive_unit
        else:
            yield transfer_or_archive_unit.first_level_choice.content_sequence
        transfer_or_archive_unit = transfer_or_archive_unit.cw_adapt_to('ITreeBase').parent()


def _inherited_rule(self, rule_type):
    """Return the rule entity of the given type, defined on this unit or in its nearest parent
    possible defining it, or None if no matching rule has been found.
    """
    assert rule_type in RULE_TYPES
    rtype = 'reverse_seda_{}_rule'.format(rule_type)
    for rule_holder in _climb_rule_holders(self):
        if getattr(rule_holder, rtype):
            return getattr(rule_holder, rtype)[0]
    return None


class SEDAArchiveTransfer(generated.SEDAArchiveTransfer):

    def dc_title(self):
        return self.title

    @property
    def formats_compat(self):
        if self.compat_list is None:
            return set()
        return set(self.compat_list.split(', '))

    @property
    def archival_agreement(self):
        if self.reverse_seda_archival_agreement:
            return self.reverse_seda_archival_agreement[0]
        return None

    @property
    def comments(self):
        return self.reverse_seda_comment

    @property
    def archive_units(self):
        return self.reverse_seda_archive_unit

    @property
    def physical_data_objects(self):
        return self.reverse_seda_physical_data_object

    @property
    def binary_data_objects(self):
        return self.reverse_seda_binary_data_object

    inherited_rule = _inherited_rule


class SEDAArchiveUnit(generated.SEDAArchiveUnit):

    def dc_title(self):
        return _extract_title(self.user_annotation)

    @property
    def is_archive_unit_ref(self):
        """Return true if this is a 'reference' archive unit, else false for 'description' archive
        unit.
        """
        return not bool(self.first_level_choice.content_sequence)

    @property
    def first_level_choice(self):
        """Return the choice element of an archive unit (SEDAAltArchiveUnitArchiveUnitRefId),
        holding either a reference or descriptive content
        """
        return self.related('seda_alt_archive_unit_archive_unit_ref_id', 'subject').one()

    inherited_rule = _inherited_rule


class SEDABinaryDataObject(generated.SEDABinaryDataObject):

    def dc_title(self):
        return _extract_title(self.user_annotation)

    @property
    def format_id(self):
        return self.reverse_seda_format_id_from[0] if self.reverse_seda_format_id_from else None

    @property
    def encoding(self):
        return self.reverse_seda_encoding_from[0] if self.reverse_seda_encoding_from else None

    @property
    def mime_type(self):
        return self.reverse_seda_mime_type_from[0] if self.reverse_seda_mime_type_from else None

    @property
    def date_created_by_application(self):
        if self.reverse_seda_date_created_by_application:
            return self.reverse_seda_date_created_by_application[0]
        return None

    @property
    def referenced_by(self):
        """Return an iterator on archive unit's content sequences referencing this data-object."""
        for ref in self.reverse_seda_data_object_reference_id:
            for seq in ref.seda_data_object_reference:
                yield seq


class SEDAPhysicalDataObject(generated.SEDAPhysicalDataObject):

    def dc_title(self):
        return _extract_title(self.user_annotation)


class SEDAAltArchiveUnitArchiveUnitRefId(generated.SEDAAltArchiveUnitArchiveUnitRefId):

    @property
    def reference(self):
        """Return the reference element for an archive unit which has no content"""
        rset = self.related('seda_archive_unit_ref_id_from', 'object')
        if rset:
            return rset.one()
        return None

    @property
    def content_sequence(self):
        """Return the sequence element holding content for an archive unit which is not a reference
        """
        rset = self.related('seda_seq_alt_archive_unit_archive_unit_ref_id_management', 'subject')
        if rset:
            return rset.one()
        return None


class SEDASeqAltArchiveUnitArchiveUnitRefIdManagement(
        generated.SEDASeqAltArchiveUnitArchiveUnitRefIdManagement):

    @property
    def title(self):
        return self.reverse_seda_title[0]

    @property
    def keywords(self):
        return self.reverse_seda_keyword

    @property
    def custodial_history_items(self):
        return self.reverse_seda_custodial_history_item

    @property
    def type(self):
        return self.reverse_seda_type_from[0] if self.reverse_seda_type_from else None

    @property
    def language(self):
        return self.reverse_seda_language_from[0] if self.reverse_seda_language_from else None

    @property
    def description_level_concept(self):
        return self.seda_description_level[0] if self.seda_description_level else None

    @property
    def description(self):
        return self.reverse_seda_description[0] if self.reverse_seda_description else None

    @property
    def start_date(self):
        return self.reverse_seda_start_date[0] if self.reverse_seda_start_date else None

    @property
    def end_date(self):
        return self.reverse_seda_end_date[0] if self.reverse_seda_end_date else None

    @property
    def system_id(self):
        return self.reverse_seda_system_id[0] if self.reverse_seda_system_id else None

    @property
    def transferring_agency_archive_unit_identifier(self):
        if self.reverse_seda_transferring_agency_archive_unit_identifier:
            return self.reverse_seda_transferring_agency_archive_unit_identifier[0]
        return None

    @property
    def originating_agency(self):
        if self.reverse_seda_originating_agency_from:
            return self.reverse_seda_originating_agency_from[0]
        return None

    inherited_rule = _inherited_rule


class SEDAKeyword(generated.SEDAKeyword):

    @property
    def reference(self):
        return (self.reverse_seda_keyword_reference_from[0]
                if self.reverse_seda_keyword_reference_from else None)

    @property
    def type(self):
        return (self.reverse_seda_keyword_type_from[0]
                if self.reverse_seda_keyword_type_from else None)


class SEDAKeywordReference(generated.SEDAKeywordReference):

    @property
    def scheme(self):
        return (self.seda_keyword_reference_to_scheme[0] if self.seda_keyword_reference_to_scheme
                else None)

    @property
    def concept(self):
        return self.seda_keyword_reference_to[0] if self.seda_keyword_reference_to else None


class SEDAKeywordType(generated.SEDAKeywordType):

    @property
    def concept(self):
        return self.seda_keyword_type_to[0] if self.seda_keyword_type_to else None


class RuleMixIn(object):

    @property
    def _rule_type(self):
        return self.cw_etype[4:-4].lower()

    @property
    def rules(self):
        return getattr(self, 'seda_seq_{0}_rule_rule'.format(self._rule_type))

    @property
    def inheritance_control(self):
        alt = getattr(self, 'seda_alt_{0}_rule_prevent_inheritance'.format(self._rule_type))
        return alt[0] if alt else None


class SEDAAccessRule(RuleMixIn, generated.SEDAAccessRule):
    pass


class SEDAAppraisalRule(RuleMixIn, generated.SEDAAppraisalRule):

    @property
    def final_action_concept(self):
        if self.seda_final_action:
            return self.seda_final_action[0]
        return None


class SEDAClassificationRule(RuleMixIn, generated.SEDAClassificationRule):
    pass


class SEDADisseminationRule(RuleMixIn, generated.SEDADisseminationRule):
    pass


class SEDAReuseRule(RuleMixIn, generated.SEDAReuseRule):
    pass


class SEDAStorageRule(RuleMixIn, generated.SEDAStorageRule):

    @property
    def final_action_concept(self):
        if self.seda_final_action:
            return self.seda_final_action[0]
        return None


class RuleRuleMixIn(object):

    @property
    def start_date(self):
        return self.reverse_seda_start_date[0] if self.reverse_seda_start_date else None

    @property
    def rule_concept(self):
        return self.seda_rule[0] if self.seda_rule else None


class SEDASeqAccessRuleRule(RuleRuleMixIn, generated.SEDASeqAccessRuleRule):
    pass


class SEDASeqAppraisalRuleRule(RuleRuleMixIn, generated.SEDASeqAppraisalRuleRule):
    pass


class SEDASeqClassificationRuleRule(RuleRuleMixIn, generated.SEDASeqClassificationRuleRule):
    pass


class SEDASeqDisseminationRuleRule(RuleRuleMixIn, generated.SEDASeqDisseminationRuleRule):
    pass


class SEDASeqReuseRuleRule(RuleRuleMixIn, generated.SEDASeqReuseRuleRule):
    pass


class SEDASeqStorageRuleRule(RuleRuleMixIn, generated.SEDASeqStorageRuleRule):
    pass


class SEDAFormatId(generated.SEDAFormatId):

    @property
    def concept(self):
        return self.seda_format_id_to[0] if self.seda_format_id_to else None


class SEDAEncoding(generated.SEDAEncoding):

    @property
    def concept(self):
        return self.seda_encoding_to[0] if self.seda_encoding_to else None


class SEDAMimeType(generated.SEDAMimeType):

    @property
    def concept(self):
        return self.seda_mime_type_to[0] if self.seda_mime_type_to else None


class SEDAType(generated.SEDAType):

    @property
    def concept(self):
        return self.seda_type_to[0] if self.seda_type_to else None


class SEDALanguage(generated.SEDALanguage):

    @property
    def concept(self):
        return self.seda_language_to[0] if self.seda_language_to else None


class SEDAOriginatingAgency(generated.SEDAOriginatingAgency):

    @property
    def agency(self):
        return self.seda_originating_agency_to[0] if self.seda_originating_agency_to else None


class SEDACustodialHistoryItem(generated.SEDACustodialHistoryItem):

    @property
    def when(self):
        return self.reverse_seda_when[0] if self.reverse_seda_when else None



# copyright 2016-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""THIS FILE IS GENERATED FROM SEDA 2.0 XSD FILES, DO NOT EDIT"""

from cubicweb.entities import AnyEntity, fetch_config


class SEDAAnyEntity(AnyEntity):
    __abstract__ = True
    value_attr = None

    def dc_title(self):
        if self.value_attr is None:
            return self.dc_type()
        return self.printable_value(self.value_attr)


class SEDAArchiveTransfer(SEDAAnyEntity):
    __regid__ = 'SEDAArchiveTransfer'
    fetch_attrs, cw_fetch_order = fetch_config(['title', 'user_annotation'])
    value_attr = None

class SEDAComment(SEDAAnyEntity):
    __regid__ = 'SEDAComment'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'comment', 'user_annotation'])
    value_attr = 'comment'

class SEDASignature(SEDAAnyEntity):
    __regid__ = 'SEDASignature'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAArchivalAgreement(SEDAAnyEntity):
    __regid__ = 'SEDAArchivalAgreement'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'archival_agreement', 'user_annotation'])
    value_attr = 'archival_agreement'

class SEDARelatedTransferReference(SEDAAnyEntity):
    __regid__ = 'SEDARelatedTransferReference'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDATransferRequestReplyIdentifier(SEDAAnyEntity):
    __regid__ = 'SEDATransferRequestReplyIdentifier'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAMessageDigestAlgorithmCodeListVersion(SEDAAnyEntity):
    __regid__ = 'SEDAMessageDigestAlgorithmCodeListVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAMimeTypeCodeListVersion(SEDAAnyEntity):
    __regid__ = 'SEDAMimeTypeCodeListVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAEncodingCodeListVersion(SEDAAnyEntity):
    __regid__ = 'SEDAEncodingCodeListVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAFileFormatCodeListVersion(SEDAAnyEntity):
    __regid__ = 'SEDAFileFormatCodeListVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDACompressionAlgorithmCodeListVersion(SEDAAnyEntity):
    __regid__ = 'SEDACompressionAlgorithmCodeListVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDADataObjectVersionCodeListVersion(SEDAAnyEntity):
    __regid__ = 'SEDADataObjectVersionCodeListVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAStorageRuleCodeListVersion(SEDAAnyEntity):
    __regid__ = 'SEDAStorageRuleCodeListVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAppraisalRuleCodeListVersion(SEDAAnyEntity):
    __regid__ = 'SEDAAppraisalRuleCodeListVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAccessRuleCodeListVersion(SEDAAnyEntity):
    __regid__ = 'SEDAAccessRuleCodeListVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDADisseminationRuleCodeListVersion(SEDAAnyEntity):
    __regid__ = 'SEDADisseminationRuleCodeListVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAReuseRuleCodeListVersion(SEDAAnyEntity):
    __regid__ = 'SEDAReuseRuleCodeListVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAClassificationRuleCodeListVersion(SEDAAnyEntity):
    __regid__ = 'SEDAClassificationRuleCodeListVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAcquisitionInformationCodeListVersion(SEDAAnyEntity):
    __regid__ = 'SEDAAcquisitionInformationCodeListVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDARelationshipCodeListVersion(SEDAAnyEntity):
    __regid__ = 'SEDARelationshipCodeListVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDABinaryDataObject(SEDAAnyEntity):
    __regid__ = 'SEDABinaryDataObject'
    fetch_attrs, cw_fetch_order = fetch_config(['filename', 'user_cardinality', 'user_annotation'])
    value_attr = None

class SEDAPhysicalDataObject(SEDAAnyEntity):
    __regid__ = 'SEDAPhysicalDataObject'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'user_annotation'])
    value_attr = None

class SEDARelationship(SEDAAnyEntity):
    __regid__ = 'SEDARelationship'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDADataObjectVersion(SEDAAnyEntity):
    __regid__ = 'SEDADataObjectVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAltBinaryDataObjectAttachment(SEDAAnyEntity):
    __regid__ = 'SEDAAltBinaryDataObjectAttachment'
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None

class SEDACompressed(SEDAAnyEntity):
    __regid__ = 'SEDACompressed'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'compressed', 'user_annotation'])
    value_attr = 'compressed'

class SEDAArchiveUnit(SEDAAnyEntity):
    __regid__ = 'SEDAArchiveUnit'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'user_annotation'])
    value_attr = None

class SEDAServiceLevel(SEDAAnyEntity):
    __regid__ = 'SEDAServiceLevel'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'service_level', 'user_annotation'])
    value_attr = 'service_level'

class SEDAAcquisitionInformation(SEDAAnyEntity):
    __regid__ = 'SEDAAcquisitionInformation'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDALegalStatus(SEDAAnyEntity):
    __regid__ = 'SEDALegalStatus'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAOriginatingAgencyIdentifier(SEDAAnyEntity):
    __regid__ = 'SEDAOriginatingAgencyIdentifier'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'originating_agency_identifier', 'user_annotation'])
    value_attr = 'originating_agency_identifier'

class SEDASubmissionAgencyIdentifier(SEDAAnyEntity):
    __regid__ = 'SEDASubmissionAgencyIdentifier'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'submission_agency_identifier', 'user_annotation'])
    value_attr = 'submission_agency_identifier'

class SEDAStorageRule(SEDAAnyEntity):
    __regid__ = 'SEDAStorageRule'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAppraisalRule(SEDAAnyEntity):
    __regid__ = 'SEDAAppraisalRule'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAccessRule(SEDAAnyEntity):
    __regid__ = 'SEDAAccessRule'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDADisseminationRule(SEDAAnyEntity):
    __regid__ = 'SEDADisseminationRule'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAReuseRule(SEDAAnyEntity):
    __regid__ = 'SEDAReuseRule'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAClassificationRule(SEDAAnyEntity):
    __regid__ = 'SEDAClassificationRule'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'classification_owner', 'user_annotation'])
    value_attr = 'classification_owner'

class SEDANeedAuthorization(SEDAAnyEntity):
    __regid__ = 'SEDANeedAuthorization'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'need_authorization', 'user_annotation'])
    value_attr = 'need_authorization'

class SEDAAltArchiveUnitArchiveUnitRefId(SEDAAnyEntity):
    __regid__ = 'SEDAAltArchiveUnitArchiveUnitRefId'
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None

class SEDAAttachment(SEDAAnyEntity):
    __regid__ = 'SEDAAttachment'
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None

class SEDAUri(SEDAAnyEntity):
    __regid__ = 'SEDAUri'
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None

class SEDAFormatLitteral(SEDAAnyEntity):
    __regid__ = 'SEDAFormatLitteral'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'format_litteral', 'user_annotation'])
    value_attr = 'format_litteral'

class SEDAMimeType(SEDAAnyEntity):
    __regid__ = 'SEDAMimeType'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAFormatId(SEDAAnyEntity):
    __regid__ = 'SEDAFormatId'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAEncoding(SEDAAnyEntity):
    __regid__ = 'SEDAEncoding'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDACreatingApplicationName(SEDAAnyEntity):
    __regid__ = 'SEDACreatingApplicationName'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'creating_application_name', 'user_annotation'])
    value_attr = 'creating_application_name'

class SEDACreatingApplicationVersion(SEDAAnyEntity):
    __regid__ = 'SEDACreatingApplicationVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'creating_application_version', 'user_annotation'])
    value_attr = 'creating_application_version'

class SEDADateCreatedByApplication(SEDAAnyEntity):
    __regid__ = 'SEDADateCreatedByApplication'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDACreatingOs(SEDAAnyEntity):
    __regid__ = 'SEDACreatingOs'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'creating_os', 'user_annotation'])
    value_attr = 'creating_os'

class SEDACreatingOsVersion(SEDAAnyEntity):
    __regid__ = 'SEDACreatingOsVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'creating_os_version', 'user_annotation'])
    value_attr = 'creating_os_version'

class SEDALastModified(SEDAAnyEntity):
    __regid__ = 'SEDALastModified'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAWidth(SEDAAnyEntity):
    __regid__ = 'SEDAWidth'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAHeight(SEDAAnyEntity):
    __regid__ = 'SEDAHeight'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDADepth(SEDAAnyEntity):
    __regid__ = 'SEDADepth'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAShape(SEDAAnyEntity):
    __regid__ = 'SEDAShape'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDADiameter(SEDAAnyEntity):
    __regid__ = 'SEDADiameter'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDALength(SEDAAnyEntity):
    __regid__ = 'SEDALength'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAThickness(SEDAAnyEntity):
    __regid__ = 'SEDAThickness'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAWeight(SEDAAnyEntity):
    __regid__ = 'SEDAWeight'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDANumberOfPage(SEDAAnyEntity):
    __regid__ = 'SEDANumberOfPage'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAArchiveUnitRefId(SEDAAnyEntity):
    __regid__ = 'SEDAArchiveUnitRefId'
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None

class SEDASeqAltArchiveUnitArchiveUnitRefIdManagement(SEDAAnyEntity):
    __regid__ = 'SEDASeqAltArchiveUnitArchiveUnitRefIdManagement'
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None

class SEDASeqStorageRuleRule(SEDAAnyEntity):
    __regid__ = 'SEDASeqStorageRuleRule'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAltStorageRulePreventInheritance(SEDAAnyEntity):
    __regid__ = 'SEDAAltStorageRulePreventInheritance'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDASeqAppraisalRuleRule(SEDAAnyEntity):
    __regid__ = 'SEDASeqAppraisalRuleRule'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAltAppraisalRulePreventInheritance(SEDAAnyEntity):
    __regid__ = 'SEDAAltAppraisalRulePreventInheritance'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDASeqAccessRuleRule(SEDAAnyEntity):
    __regid__ = 'SEDASeqAccessRuleRule'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAltAccessRulePreventInheritance(SEDAAnyEntity):
    __regid__ = 'SEDAAltAccessRulePreventInheritance'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDASeqDisseminationRuleRule(SEDAAnyEntity):
    __regid__ = 'SEDASeqDisseminationRuleRule'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAltDisseminationRulePreventInheritance(SEDAAnyEntity):
    __regid__ = 'SEDAAltDisseminationRulePreventInheritance'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDASeqReuseRuleRule(SEDAAnyEntity):
    __regid__ = 'SEDASeqReuseRuleRule'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAltReuseRulePreventInheritance(SEDAAnyEntity):
    __regid__ = 'SEDAAltReuseRulePreventInheritance'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDASeqClassificationRuleRule(SEDAAnyEntity):
    __regid__ = 'SEDASeqClassificationRuleRule'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAltClassificationRulePreventInheritance(SEDAAnyEntity):
    __regid__ = 'SEDAAltClassificationRulePreventInheritance'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAClassificationReassessingDate(SEDAAnyEntity):
    __regid__ = 'SEDAClassificationReassessingDate'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDANeedReassessingAuthorization(SEDAAnyEntity):
    __regid__ = 'SEDANeedReassessingAuthorization'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'need_reassessing_authorization', 'user_annotation'])
    value_attr = 'need_reassessing_authorization'

class SEDADataObjectReference(SEDAAnyEntity):
    __regid__ = 'SEDADataObjectReference'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDATitle(SEDAAnyEntity):
    __regid__ = 'SEDATitle'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'title', 'user_annotation'])
    value_attr = 'title'

class SEDAFilePlanPosition(SEDAAnyEntity):
    __regid__ = 'SEDAFilePlanPosition'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'file_plan_position', 'user_annotation'])
    value_attr = 'file_plan_position'

class SEDASystemId(SEDAAnyEntity):
    __regid__ = 'SEDASystemId'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAOriginatingSystemId(SEDAAnyEntity):
    __regid__ = 'SEDAOriginatingSystemId'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAArchivalAgencyArchiveUnitIdentifier(SEDAAnyEntity):
    __regid__ = 'SEDAArchivalAgencyArchiveUnitIdentifier'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'archival_agency_archive_unit_identifier', 'user_annotation'])
    value_attr = 'archival_agency_archive_unit_identifier'

class SEDAOriginatingAgencyArchiveUnitIdentifier(SEDAAnyEntity):
    __regid__ = 'SEDAOriginatingAgencyArchiveUnitIdentifier'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'originating_agency_archive_unit_identifier', 'user_annotation'])
    value_attr = 'originating_agency_archive_unit_identifier'

class SEDATransferringAgencyArchiveUnitIdentifier(SEDAAnyEntity):
    __regid__ = 'SEDATransferringAgencyArchiveUnitIdentifier'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'transferring_agency_archive_unit_identifier', 'user_annotation'])
    value_attr = 'transferring_agency_archive_unit_identifier'

class SEDADescription(SEDAAnyEntity):
    __regid__ = 'SEDADescription'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'description', 'user_annotation'])
    value_attr = 'description'

class SEDAType(SEDAAnyEntity):
    __regid__ = 'SEDAType'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDADocumentType(SEDAAnyEntity):
    __regid__ = 'SEDADocumentType'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'document_type', 'user_annotation'])
    value_attr = 'document_type'

class SEDALanguage(SEDAAnyEntity):
    __regid__ = 'SEDALanguage'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDADescriptionLanguage(SEDAAnyEntity):
    __regid__ = 'SEDADescriptionLanguage'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAStatus(SEDAAnyEntity):
    __regid__ = 'SEDAStatus'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'status', 'user_annotation'])
    value_attr = 'status'

class SEDAVersion(SEDAAnyEntity):
    __regid__ = 'SEDAVersion'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'version', 'user_annotation'])
    value_attr = 'version'

class SEDATag(SEDAAnyEntity):
    __regid__ = 'SEDATag'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'tag', 'user_annotation'])
    value_attr = 'tag'

class SEDAKeyword(SEDAAnyEntity):
    __regid__ = 'SEDAKeyword'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'keyword_content', 'user_annotation'])
    value_attr = 'keyword_content'

class SEDAOriginatingAgency(SEDAAnyEntity):
    __regid__ = 'SEDAOriginatingAgency'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDASubmissionAgency(SEDAAnyEntity):
    __regid__ = 'SEDASubmissionAgency'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAuthorizedAgent(SEDAAnyEntity):
    __regid__ = 'SEDAAuthorizedAgent'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAWriter(SEDAAnyEntity):
    __regid__ = 'SEDAWriter'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAddressee(SEDAAnyEntity):
    __regid__ = 'SEDAAddressee'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDARecipient(SEDAAnyEntity):
    __regid__ = 'SEDARecipient'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDASource(SEDAAnyEntity):
    __regid__ = 'SEDASource'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'source', 'user_annotation'])
    value_attr = 'source'

class SEDACreatedDate(SEDAAnyEntity):
    __regid__ = 'SEDACreatedDate'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDATransactedDate(SEDAAnyEntity):
    __regid__ = 'SEDATransactedDate'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAcquiredDate(SEDAAnyEntity):
    __regid__ = 'SEDAAcquiredDate'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDASentDate(SEDAAnyEntity):
    __regid__ = 'SEDASentDate'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAReceivedDate(SEDAAnyEntity):
    __regid__ = 'SEDAReceivedDate'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDARegisteredDate(SEDAAnyEntity):
    __regid__ = 'SEDARegisteredDate'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAStartDate(SEDAAnyEntity):
    __regid__ = 'SEDAStartDate'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAEndDate(SEDAAnyEntity):
    __regid__ = 'SEDAEndDate'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAEvent(SEDAAnyEntity):
    __regid__ = 'SEDAEvent'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAPreventInheritance(SEDAAnyEntity):
    __regid__ = 'SEDAPreventInheritance'
    fetch_attrs, cw_fetch_order = fetch_config(['prevent_inheritance'])
    value_attr = 'prevent_inheritance'

class SEDARefNonRuleId(SEDAAnyEntity):
    __regid__ = 'SEDARefNonRuleId'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDACustodialHistoryItem(SEDAAnyEntity):
    __regid__ = 'SEDACustodialHistoryItem'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'custodial_history_item', 'user_annotation'])
    value_attr = 'custodial_history_item'

class SEDACustodialHistoryFile(SEDAAnyEntity):
    __regid__ = 'SEDACustodialHistoryFile'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAKeywordReference(SEDAAnyEntity):
    __regid__ = 'SEDAKeywordReference'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAKeywordType(SEDAAnyEntity):
    __regid__ = 'SEDAKeywordType'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDASpatial(SEDAAnyEntity):
    __regid__ = 'SEDASpatial'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'spatial', 'user_annotation'])
    value_attr = 'spatial'

class SEDATemporal(SEDAAnyEntity):
    __regid__ = 'SEDATemporal'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'temporal', 'user_annotation'])
    value_attr = 'temporal'

class SEDAJuridictional(SEDAAnyEntity):
    __regid__ = 'SEDAJuridictional'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality', 'juridictional', 'user_annotation'])
    value_attr = 'juridictional'

class SEDAIsVersionOf(SEDAAnyEntity):
    __regid__ = 'SEDAIsVersionOf'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAReplaces(SEDAAnyEntity):
    __regid__ = 'SEDAReplaces'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDARequires(SEDAAnyEntity):
    __regid__ = 'SEDARequires'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAIsPartOf(SEDAAnyEntity):
    __regid__ = 'SEDAIsPartOf'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAReferences(SEDAAnyEntity):
    __regid__ = 'SEDAReferences'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAEventIdentifier(SEDAAnyEntity):
    __regid__ = 'SEDAEventIdentifier'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAEventType(SEDAAnyEntity):
    __regid__ = 'SEDAEventType'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAEventDetail(SEDAAnyEntity):
    __regid__ = 'SEDAEventDetail'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAGpsVersionID(SEDAAnyEntity):
    __regid__ = 'SEDAGpsVersionID'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAGpsAltitude(SEDAAnyEntity):
    __regid__ = 'SEDAGpsAltitude'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAGpsAltitudeRef(SEDAAnyEntity):
    __regid__ = 'SEDAGpsAltitudeRef'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAGpsLatitude(SEDAAnyEntity):
    __regid__ = 'SEDAGpsLatitude'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAGpsLatitudeRef(SEDAAnyEntity):
    __regid__ = 'SEDAGpsLatitudeRef'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAGpsLongitude(SEDAAnyEntity):
    __regid__ = 'SEDAGpsLongitude'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAGpsLongitudeRef(SEDAAnyEntity):
    __regid__ = 'SEDAGpsLongitudeRef'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAGpsDateStamp(SEDAAnyEntity):
    __regid__ = 'SEDAGpsDateStamp'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAwhen(SEDAAnyEntity):
    __regid__ = 'SEDAwhen'
    fetch_attrs, cw_fetch_order = fetch_config(['user_cardinality'])
    value_attr = None

class SEDAAltIsVersionOfArchiveUnitRefId(SEDAAnyEntity):
    __regid__ = 'SEDAAltIsVersionOfArchiveUnitRefId'
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None

class SEDAAltReplacesArchiveUnitRefId(SEDAAnyEntity):
    __regid__ = 'SEDAAltReplacesArchiveUnitRefId'
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None

class SEDAAltRequiresArchiveUnitRefId(SEDAAnyEntity):
    __regid__ = 'SEDAAltRequiresArchiveUnitRefId'
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None

class SEDAAltIsPartOfArchiveUnitRefId(SEDAAnyEntity):
    __regid__ = 'SEDAAltIsPartOfArchiveUnitRefId'
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None

class SEDAAltReferencesArchiveUnitRefId(SEDAAnyEntity):
    __regid__ = 'SEDAAltReferencesArchiveUnitRefId'
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None

class SEDARepositoryArchiveUnitPID(SEDAAnyEntity):
    __regid__ = 'SEDARepositoryArchiveUnitPID'
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None

class SEDARepositoryObjectPID(SEDAAnyEntity):
    __regid__ = 'SEDARepositoryObjectPID'
    fetch_attrs, cw_fetch_order = fetch_config([])
    value_attr = None

CHOICE_RTYPE = {
  'SEDAAltAccessRulePreventInheritance': [
    [
      'seda_prevent_inheritance',
      'object'
    ],
    [
      'seda_ref_non_rule_id_from',
      'object'
    ]
  ],
  'SEDAAltAppraisalRulePreventInheritance': [
    [
      'seda_prevent_inheritance',
      'object'
    ],
    [
      'seda_ref_non_rule_id_from',
      'object'
    ]
  ],
  'SEDAAltArchiveUnitArchiveUnitRefId': [
    [
      'seda_archive_unit_ref_id_from',
      'object'
    ],
    [
      'seda_seq_alt_archive_unit_archive_unit_ref_id_management',
      'subject'
    ]
  ],
  'SEDAAltBinaryDataObjectAttachment': [
    [
      'seda_attachment',
      'object'
    ],
    [
      'seda_uri',
      'object'
    ]
  ],
  'SEDAAltClassificationRulePreventInheritance': [
    [
      'seda_prevent_inheritance',
      'object'
    ],
    [
      'seda_ref_non_rule_id_from',
      'object'
    ]
  ],
  'SEDAAltDisseminationRulePreventInheritance': [
    [
      'seda_prevent_inheritance',
      'object'
    ],
    [
      'seda_ref_non_rule_id_from',
      'object'
    ]
  ],
  'SEDAAltIsPartOfArchiveUnitRefId': [
    [
      'seda_archive_unit_ref_id_from',
      'object'
    ],
    [
      'seda_repository_archive_unit_pid',
      'object'
    ],
    [
      'seda_repository_object_pid',
      'object'
    ],
    [
      'seda_data_object_reference',
      'object'
    ]
  ],
  'SEDAAltIsVersionOfArchiveUnitRefId': [
    [
      'seda_archive_unit_ref_id_from',
      'object'
    ],
    [
      'seda_data_object_reference',
      'object'
    ],
    [
      'seda_repository_archive_unit_pid',
      'object'
    ],
    [
      'seda_repository_object_pid',
      'object'
    ]
  ],
  'SEDAAltReferencesArchiveUnitRefId': [
    [
      'seda_archive_unit_ref_id_from',
      'object'
    ],
    [
      'seda_repository_archive_unit_pid',
      'object'
    ],
    [
      'seda_repository_object_pid',
      'object'
    ],
    [
      'seda_data_object_reference',
      'object'
    ]
  ],
  'SEDAAltReplacesArchiveUnitRefId': [
    [
      'seda_archive_unit_ref_id_from',
      'object'
    ],
    [
      'seda_repository_archive_unit_pid',
      'object'
    ],
    [
      'seda_repository_object_pid',
      'object'
    ],
    [
      'seda_data_object_reference',
      'object'
    ]
  ],
  'SEDAAltRequiresArchiveUnitRefId': [
    [
      'seda_archive_unit_ref_id_from',
      'object'
    ],
    [
      'seda_repository_archive_unit_pid',
      'object'
    ],
    [
      'seda_repository_object_pid',
      'object'
    ],
    [
      'seda_data_object_reference',
      'object'
    ]
  ],
  'SEDAAltReuseRulePreventInheritance': [
    [
      'seda_prevent_inheritance',
      'object'
    ],
    [
      'seda_ref_non_rule_id_from',
      'object'
    ]
  ],
  'SEDAAltStorageRulePreventInheritance': [
    [
      'seda_prevent_inheritance',
      'object'
    ],
    [
      'seda_ref_non_rule_id_from',
      'object'
    ]
  ]
}



# copyright 2016-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-seda ITreeBase adapters"""

from cubicweb.predicates import is_instance
from cubicweb.view import EntityAdapter

from . import simplified_profile


def parent_archive_unit(entity):
    """Return the first encountered parent which is an ArchiveUnit"""
    while entity.cw_etype != 'SEDAArchiveUnit':
        entity = entity.cw_adapt_to('IContained').parent
    return entity


class IContainedToITreeBase(EntityAdapter):
    """Map IContained adapters to ITreeBase, additionaly configured with a list of relations leading to
    contained's children.

    ITreeBase is a simplified version of cubicweb's ITree.
    """

    __regid__ = 'ITreeBase'
    __abstract__ = True

    _children_relations = []  # list of (relation type, role) to get entity's children

    def iterchildren(self):
        """Return an iterator over the item's children."""
        for rel, role in self._children_relations:
            for child in self.entity.related(rel, role, entities=True):
                yield child

    def is_leaf(self):
        """Returns True if the entity does not have any children."""
        try:
            next(iter(self.iterchildren()))
            return False
        except StopIteration:
            return True

    def parent(self):
        return self.entity.cw_adapt_to('IContained').parent

    def iterancestors(self):
        """Return an iterator on the ancestors of the entity."""
        def _uptoroot(self):
            curr = self
            while True:
                curr = curr.parent()
                if curr is None:
                    break
                yield curr
                curr = curr.cw_adapt_to('ITreeBase')
        return _uptoroot(self)

    iterparents = iterancestors  # Compat. with CubicWeb's ITree.


class ITreeBaseArchiveUnitAdapter(IContainedToITreeBase):
    """Adapt ArchiveUnit entities to ITreeBase."""

    __select__ = is_instance('SEDAArchiveUnit')

    _children_relations = [('seda_archive_unit', 'object')]

    def parent(self):
        parent = self.entity.cw_adapt_to('IContained').parent
        if parent is None:
            return None
        if parent.cw_etype == 'SEDAArchiveTransfer':
            return parent
        return parent_archive_unit(parent)

    def iterchildren(self):
        seq = self.entity.first_level_choice.content_sequence
        if seq is None:
            # 'reference' archive unit
            return
        for rel, role in self._children_relations:
            for child in seq.related(rel, role, entities=True):
                yield child


class ITreeBaseSimplifiedArchiveUnitAdapter(ITreeBaseArchiveUnitAdapter):

    __select__ = ITreeBaseArchiveUnitAdapter.__select__ & simplified_profile()

    _children_relations = [('seda_archive_unit', 'object')]

    def iterchildren(self):
        for child in super(ITreeBaseSimplifiedArchiveUnitAdapter, self).iterchildren():
            yield child
        seq = self.entity.first_level_choice.content_sequence
        assert seq is not None, self.entity  # can't be None in simplified profile
        for do in self._cw.execute(
                'Any DO, DOUA ORDERBY DOUA WHERE DO user_annotation DOUA, '
                'REF seda_data_object_reference_id DO, '
                'REF seda_data_object_reference SEQ, SEQ eid %(x)s',
                {'x': seq.eid}).entities():
            yield do


class ITreeBaseDataObjectAdapter(IContainedToITreeBase):
    """Adapt BinaryDataObject and PhysicalDataObject entities to ITreeBase."""

    __select__ = is_instance('SEDABinaryDataObject', 'SEDAPhysicalDataObject')


class ITreeBaseSimplifiedDataObjectAdapter(ITreeBaseDataObjectAdapter):

    __select__ = ITreeBaseDataObjectAdapter.__select__ & simplified_profile()

    def parent(self):
        return parent_archive_unit(self.entity.reverse_seda_data_object_reference_id[0])


class ITreeBaseArchiveTransferAdapter(IContainedToITreeBase):
    """Adapt ArchiveTransfer entities to ITreeBase."""

    __select__ = is_instance('SEDAArchiveTransfer')

    @property
    def _children_relations(self):
        if self.entity.simplified_profile:
            return [('seda_archive_unit', 'object')]
        else:
            return [('seda_binary_data_object', 'object'),
                    ('seda_physical_data_object', 'object'),
                    ('seda_archive_unit', 'object')]

    def parent(self):
        return None



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Extract concepts scheme as Linked CSV from XSD file.

XSD parsing is done using generateDS, which has been copied into the `gends` directory (only the
used bits).
"""
import csv
import os.path as osp

from xsd import seda_xsd, un_camel_case

import sys
sys.exit('broken until pyxst stop using a set for textual_content_values')


def generate_csv(output_dir):
    xschema = seda_xsd()
    for xstype, attr_or_idx in (
            ('Diameter', 'unit'),
            ('Weight', 'unit'),
            ('FinalAction', 0),
            ('FinalAction', 1),
            ('KeywordType', 0),
            ('DescriptionLevel', 0)):
        xselements = xschema.elts_index['{fr:gouv:culture:archivesdefrance:seda:v2.0}' + xstype]
        if isinstance(attr_or_idx, int):
            xselement = xselements[attr_or_idx]
        else:
            assert len(xselements) == 1
            xselement = xselements[0].attributes[attr_or_idx].target
        with open(osp.join(output_dir, un_camel_case(xstype) + '.csv'), 'w') as stream:
            writer = csv.writer(stream, delimiter=';', quotechar='"', quoting=csv.QUOTE_MINIMAL)
            _GENERATORS['{0}_{1}'.format(xstype, attr_or_idx)](writer,
                                                               xselement.textual_content_values)


def _generate_pref_alt(writer, values):
    writer.writerow(['#', '$id', 'label', 'alt label'])
    writer.writerow(['type', 'url', 'string', 'string'])
    writer.writerow(['lang', '', 'en', 'en'])
    writer.writerow(['url', 'skos:Concept', 'skos:prefLabel', 'skos:altLabel'])

    def writerow(row):
        assert 1 <= len(row) <= 2, len(row)
        if len(row) == 1:
            row.append('')
        writer.writerow(['', ''] + row)

    row = []
    for value in values:
        if not row or value.isupper():
            row.append(value)
        else:
            writerow(row)
            row = [value]
    writerow(row)


def _generate_pref(writer, values):
    writer.writerow(['#', '$id', 'label'])
    writer.writerow(['type', 'url', 'string'])
    writer.writerow(['lang', '', 'en'])
    writer.writerow(['url', 'skos:Concept', 'skos:prefLabel'])
    for value in values:
        writer.writerow(['', '', value])


_GENERATORS = {
    'Diameter_unit': _generate_pref_alt,
    'Weight_unit': _generate_pref_alt,
    'FinalAction_0': _generate_pref,
    'FinalAction_1': _generate_pref,
    'KeywordType_0': _generate_pref,
    'DescriptionLevel_0': _generate_pref,
}

if __name__ == '__main__':
    import sys
    if len(sys.argv) == 2:
        output_dir = sys.argv[1]
    elif len(sys.argv) < 2:
        output_dir = '.'
    else:
        raise Exception('too much arguments')
    generate_csv(output_dir)



# pylint: disable=W0622
"""cubicweb-seda application packaging information"""

modname = 'seda'
distname = 'cubicweb-seda'

numversion = (0, 10, 2)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
author = 'LOGILAB S.A. (Paris, FRANCE)'
author_email = 'contact@logilab.fr'
description = 'Data Exchange Standard for Archival'
web = 'http://www.cubicweb.org/project/%s' % distname

__depends__ = {
    'cubicweb': '>= 3.24, < 3.25',
    'six': '>= 1.4.0',
    'cubicweb-eac': None,
    'cubicweb-skos': '>= 0.12.1',
    'cubicweb-compound': '>= 0.6',
    'cubicweb-relationwidget': '>= 0.4',
    'cubicweb-squareui': None,
    'pyxst': None,
    'rdflib': '>= 4.1',
}
__recommends__ = {}

classifiers = [
    'Environment :: Web Environment',
    'Framework :: CubicWeb',
    'Programming Language :: Python',
    'Programming Language :: JavaScript',
]



# copyright 2016-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Generate CubicWeb's uicfg rules from XSD file.

XSD parsing is done using generateDS, which has been copied into the `gends` directory (only the
used bits).
"""

from six import text_type

from cubicweb import neg_role, _

from xsd import XSDM_MAPPING
from xsd2yams import CodeGenerator


FIRST_LEVEL_ETYPES = set(('SEDAArchiveTransfer',
                          'SEDABinaryDataObject', 'SEDAPhysicalDataObject',
                          'SEDAArchiveUnit'))

RTYPES_IN_TAB = set((
    'seda_binary_data_object',
    'seda_physical_data_object',
    'seda_related_transfer_reference',
    'seda_archive_unit',
    'seda_relationship',
    'seda_storage_rule',
    'seda_appraisal_rule',
    'seda_access_rule',
    'seda_dissemination_rule',
    'seda_reuse_rule',
    'seda_classification_rule',
    'seda_need_authorization',
    'seda_restriction_rule_id_ref',
    'seda_restriction_value',
    'seda_restriction_end_date',
))
for element_name in ('CodeListVersions', 'FormatIdentification', 'FileInfo', 'PhysicalDimensions',
                     'Gps', 'RelatedObjectReference', 'CustodialHistory', 'Coverage'):
    for rtype, role, path in XSDM_MAPPING.iter_rtype_role(element_name):
        RTYPES_IN_TAB.add(rtype)


class UICFGGenerator(CodeGenerator):
    """UICFG rules generator"""

    rtags_info = {
        'actionbox_appearsin_addmenu': {
            'shortname': 'abaa',
            'subject': "abaa.tag_subject_of(('*', '{rtype}', '*'), {value})",
            'object': "abaa.tag_object_of(('*', '{rtype}', '*'), {value})",
        },
        'autoform_section': {
            'shortname': 'afs',
            'subject': "afs.tag_subject_of(('*', '{rtype}', '*'), 'main', '{value}')",
            'object': "afs.tag_object_of(('*', '{rtype}', '*'), 'main', '{value}')",
        },
        'autoform_field_kwargs': {
            'shortname': 'affk',
            'subject': "affk.tag_subject_of(('*', '{rtype}', '*'), {value})",
            'object': "affk.tag_object_of(('*', '{rtype}', '*'), {value})",
        },
        'primaryview_section': {
            'shortname': 'pvs',
            'subject': "pvs.tag_subject_of(('*', '{rtype}', '*'), '{value}')",
            'object': "pvs.tag_object_of(('*', '{rtype}', '*'), '{value}')",
        },
        'reledit_ctrl': {
            'shortname': 'rec',
            'subject': "rec.tag_subject_of(('*', '{rtype}', '*'), "
            "{{'novalue_label': '{value}'}})",
            'object': "rec.tag_object_of(('*', '{rtype}', '*'), "
            "{{'novalue_label': '{value}'}})",
        },
    }

    def _generate(self, mapping, stream):
        self._processed = set()
        stream.write('from cubicweb.web import formwidgets as fw\n')
        stream.write('from cubicweb.web.views import uicfg\n\n')
        # indexview_etype_section configuration
        stream.write('ives = uicfg.indexview_etype_section\n')
        all_etypes = set()
        for mapping_element in mapping.ordered:
            etypes = self._callback('etypes_for', mapping_element)
            all_etypes.update(set(etypes))
        all_etypes.remove('SEDAArchiveTransfer')
        for etype in all_etypes:
            stream.write("ives['{0}'] = 'subobject'\n".format(etype))
        # autoform / primary view relation rtags
        for rtag, rtag_info in sorted(self.rtags_info.items()):
            alias = rtag_info['shortname']
            stream.write('\n\n{0} = uicfg.{1}\n'.format(rtag_info['shortname'], rtag))
            for mapping_element in mapping.ordered:
                for rtype, role, value in self._callback(rtag + '_for', mapping_element):
                    template = rtag_info[role]
                    stream.write(template.format(**locals()) + '\n')
        # fields order
        stream.write('pvds = uicfg.primaryview_display_ctrl\n')
        for mapping_element in mapping.ordered:
            for etype, attributes in self._callback('order_for', mapping_element):
                stream.write("affk.set_fields_order('{0}', {1})\n".format(etype, attributes))
                stream.write("pvds.set_fields_order('{0}', {1})\n".format(etype, attributes))

    def etypes_for_e_type_mapping(self, mapping):
        yield mapping.etype

    def autoform_section_for_rdef_mapping(self, mapping):
        if ('afs', mapping.rtype) in self._processed:
            return
        self._processed.add(('afs', mapping.rtype))
        if mapping.rtype in RTYPES_IN_TAB:
            section = 'hidden'
            role = mapping.composite or 'subject'
        elif mapping.composite:
            section = 'inlined'
            role = mapping.composite
        else:
            section = 'attributes'
            role = 'subject'
        yield mapping.rtype, neg_role(role), 'hidden'
        yield mapping.rtype, role, section

    def autoform_field_kwargs_for_e_type_mapping(self, mapping):
        for rtype, target_etype in sorted(mapping.attributes.items()):
            if rtype == 'id':
                continue
            if target_etype == 'String' and ('affk', rtype) not in self._processed:
                self._processed.add(('affk', rtype))
                yield rtype, 'subject', {'widget': Code("fw.TextInput({'size': 80})")}
            elif target_etype == 'Boolean' and ('aff', rtype) not in self._processed:
                self._processed.add(('aff', rtype))
                yield rtype, 'subject', {'allow_none': True}

    def autoform_field_kwargs_for_rdef_mapping(self, mapping):
        if mapping.rtype.endswith('code_list_version_to'):
            yield mapping.rtype, 'subject', {'label': 'value'}

    def primaryview_section_for_rdef_mapping(self, mapping):
        if ('pvs', mapping.rtype) in self._processed:
            return
        self._processed.add(('pvs', mapping.rtype))
        if mapping.rtype in RTYPES_IN_TAB:
            section = 'hidden'
            role = mapping.composite or 'subject'
            yield mapping.rtype, neg_role(role), section
            yield mapping.rtype, role, section
        elif ('Concept' in mapping.objtypes
              or 'ConceptScheme' in mapping.objtypes
              or 'AuthorityRecord' in mapping.objtypes):
            yield mapping.rtype, 'object', 'hidden'

    def reledit_ctrl_for_e_type_mapping(self, mapping):
        for rtype, target_etype in sorted(mapping.attributes.items()):
            if rtype == 'id':
                continue
            if ('rec', rtype) not in self._processed:
                self._processed.add(('rec', rtype))
                yield rtype, 'subject', _('<no value specified>')

    def reledit_ctrl_for_rdef_mapping(self, mapping):
        if ('rec', mapping.rtype) in self._processed:
            return
        self._processed.add(('rec', mapping.rtype))
        if mapping.composite is None:
            yield mapping.rtype, 'subject', '<no value specified>'
        else:
            card = mapping.card[mapping.composite == 'object']
            if card == '1':
                yield mapping.rtype, mapping.composite, _('<no value specified>')
            elif card == '?':
                yield mapping.rtype, mapping.composite, ' '

    def actionbox_appearsin_addmenu_for_rdef_mapping(self, mapping):
        if ('abaa', mapping.rtype) in self._processed:
            return
        self._processed.add(('abaa', mapping.rtype))
        role = mapping.composite or 'subject'
        yield mapping.rtype, neg_role(role), False
        yield mapping.rtype, role, False

    def order_for_e_type_mapping(self, mapping):
        attributes = ordered_attributes(mapping)
        if len(attributes) > 1:
            yield mapping.etype, attributes


class Code(text_type):
    """Special string subclass whose repr() doesn't add quotes, for insertion of python code in a
    data structure
    """
    def __repr__(self):
        return str(self)


def ordered_attributes(mapping):
    """Given an ETypeMapping, return a list of its attributes sorted by desired order of appearance
    """
    attributes = [attr for attr in mapping.attributes if attr != 'id']
    if mapping.etype == 'SEDAArchiveTransfer':
        attributes.append('title')
        attributes.append('user_annotation')
    elif mapping.etype in ('SEDAArchiveUnit', 'SEDABinaryDataObject', 'SEDAPhysicalDataObject'):
        attributes.append('user_cardinality')
        attributes.append('user_annotation')
    elif attributes:
        if len(mapping.cards) > 1:
            attributes.insert(0, 'user_cardinality')
            attributes.append('user_annotation')
    return attributes


if __name__ == '__main__':
    UICFGGenerator.main()
